
import numpy as np

class BBoxEvaluator:
    @staticmethod
    def load_BBox_results(input_filename):
        with open(input_filename, "r", encoding="utf8") as in_file:
            all_lines = in_file.readlines()

        # assuming first line is header ...
        results_per_query = {}
        for idx, line in enumerate(all_lines[1:]):
            parts = line.strip().split(",")

            # Query id, location id, x, y, w, h, rank, score(s)
            if len(parts) >= 7:
                try:
                    query_id = parts[0].strip()
                    location_id = parts[1].strip()
                    x = int(parts[2])
                    y = int(parts[3])
                    w = int(parts[4])
                    h = int(parts[5])
                    rank = int(parts[6])
                except:
                    raise Exception("Error loading Results line #{0:d}:\n{1:s}".format(idx + 1, line))

                if query_id not in results_per_query:
                    results_per_query[query_id] = []

                results_per_query[query_id].append((rank, location_id, (x, y, w, h)))

        # ensure sorted by rank ...
        for query_id in results_per_query:
            sorted_locations = sorted(results_per_query[query_id], key=lambda x: x[0])
            results_per_query[query_id] = [(loc, box) for _, loc, box in sorted_locations]

        return results_per_query

    @staticmethod
    def load_BBox_ground_truth(input_filename):
        with open(input_filename, "r", encoding="utf8") as in_file:
            all_lines = in_file.readlines()

        # assuming first line is header ...
        gt_per_query = {}
        for idx, line in enumerate(all_lines[1:]):
            parts = line.strip().split(",")

            # query_id, location_id, x, y, w, h, relevance
            if len(parts) >= 7:
                try:
                    query_id = parts[0].strip()
                    location_id = parts[1].strip()

                    x = int(parts[2])
                    y = int(parts[3])
                    w = int(parts[4])
                    h = int(parts[5])
                    relevance = int(parts[6])
                except:
                    raise Exception("Error loading Ground Truth line #{0:d}:\n{1:s}".format(idx + 1, line))

                if not query_id in gt_per_query:
                    gt_per_query[query_id] = {}

                if not location_id in gt_per_query[query_id]:
                    gt_per_query[query_id][location_id] = []

                gt_per_query[query_id][location_id].append(((x, y, w, h), relevance))

        return gt_per_query

    @staticmethod
    def save_BBox_unjudged(output_filename, unjudged_results):
        all_lines = ["query_id, location, x, y, w, h, rank\n"]
        for query_id, loc, m_x, m_y, m_w, m_h, rank in unjudged_results:
            line = ",".join([query_id, loc, str(m_x), str(m_y), str(m_w), str(m_h), str(rank)]) + "\n"
            all_lines.append(line)

        with open(output_filename, "w", encoding="utf8") as out_file:
            out_file.writelines(all_lines)

    @staticmethod
    def get_bboxes_IOU(bbox1, bbox2):
        b1_x, b1_y, b1_w, b1_h = bbox1
        b2_x, b2_y, b2_w, b2_h = bbox2

        int_min_x = max(b1_x, b2_x)
        int_min_y = max(b1_y, b2_y)
        int_max_x = min(b1_x + b1_w, b2_x + b2_w)
        int_max_y = min(b1_y + b1_h, b2_y + b2_h)

        int_w = max(0, int_max_x - int_min_x)
        int_h = max(0, int_max_y - int_min_y)

        int_area = int_w * int_h
        union_area = (b1_w * b1_h) + (b2_w * b2_h) - int_area

        iou = int_area / union_area
        return iou


    @staticmethod
    def locate_target_rank(results_bboxes, gt_bboxes, min_match_IOU=0.5, verbose=True):
        all_query_ids = sorted(list(set(results_bboxes.keys()).union(set(gt_bboxes.keys()))))
        all_unjudged = []

        warning_msg = "Warning: No assessment for {0:s}, match #{1:d} ({2:d}, {3:d}, {4:d}, {5:d})"

        target_ranks = {}
        for query_id in all_query_ids:
            # by default, assume that target was not found ...
            target_ranks[query_id] = -1

            if not query_id in results_bboxes:
                if verbose:
                    print("Warning: {0:s} in GT but not in Results".format(query_id))

                continue

            if not query_id in gt_bboxes:
                if verbose:
                    print("Warning: {0:s} in Results but not in GT".format(query_id))

                continue

            # if reaches here, the element is in both!
            query_gt = gt_bboxes[query_id]

            # find first matching target from current list of sorted results
            for offset, (loc, m_bbox) in enumerate(results_bboxes[query_id]):
                m_x, m_y, m_w, m_h = m_bbox

                # check if location appears in ground truth ...
                if not loc in query_gt:
                    all_unjudged.append((query_id, loc, m_x, m_y, m_w, m_h, offset + 1))
                    if verbose:
                        print(warning_msg.format(query_id, offset + 1, m_x, m_y, m_w, m_h), flush=True)
                    continue

                # location is in GT, check if any of the boxes overlaps
                all_boxes_overlap = []
                for qrel_bbox, relevance in query_gt[loc]:
                    # get overlap metric
                    iou = BBoxEvaluator.get_bboxes_IOU(m_bbox, qrel_bbox)

                    # if overlap is large enough ...
                    if iou >= min_match_IOU:
                        all_boxes_overlap.append((iou, relevance))

                # check if assessment has been found in GT ...
                if len(all_boxes_overlap) == 0:
                    # no assessments found in GT for this match
                    if verbose:
                        all_unjudged.append((query_id, loc, m_x, m_y, m_w, m_h, offset + 1))
                        print(warning_msg.format(query_id, offset + 1, m_x, m_y, m_w, m_h), flush=True)
                else:
                    # sort matching assessments ...
                    all_boxes_overlap = sorted(all_boxes_overlap, key=lambda x: x[0], reverse=True)

                    # take top match (bigger overlap) ...
                    iou, relevance = all_boxes_overlap[0]

                    if relevance > 0:
                        # a target has been found!
                        target_ranks[query_id] = offset + 1
                        # stop search for this query
                        break

        return all_query_ids, target_ranks, all_unjudged

    @staticmethod
    def get_MRR(results_bboxes, gt_bboxes, min_match_IOU=0.5, verbose=True):
        all_query_ids, all_ranks, all_unjudged = BBoxEvaluator.locate_target_rank(results_bboxes, gt_bboxes,
                                                                                  min_match_IOU, verbose)

        # get the RR and then the means ...
        RR = np.zeros(len(all_query_ids))
        matched = 0
        for idx, query_id in enumerate(all_query_ids):
            if all_ranks[query_id] > 0:
                RR[idx] = 1 / all_ranks[query_id]
                matched += 1

        recall = matched / len(all_query_ids)

        overall = {
            "recall": recall,
            "mrr": RR.mean(),
            "unjudged": all_unjudged,
        }

        per_query = {
            "query_ids":  all_query_ids,
            "rr": RR,
        }

        return overall, per_query

    @staticmethod
    def display_MRR(overall, per_query):

        print("\nEvaluation Target Recall and MRR\n")
        print("Query\tMatch\tRR")

        RR =  per_query["rr"]
        output_line = "{0:s}\t{1:d}\t{2:.4f}"
        for idx, query_id in enumerate(per_query["query_ids"]):

            matched = 1 if RR[idx] > 0 else 0
            print(output_line.format(query_id, matched, RR[idx]))

        print("Recall\t{0:.4f}".format(overall["recall"]))
        print("MRR\t{0:.4f}".format(overall["mrr"]))
