
import gc
import time
import pickle
# from AccessMath.annotation.keyframe_annotation import KeyFrameAnnotation
# from AccessMath.data.lecture_info import LectureInfo

from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.indexing.graph_index_entry import GraphIndexEntry
from TangentV.indexing.graph_raw_pair import GraphRawPair
from TangentV.indexing.index_image import IndexImage
from TangentV.retrieval.graph_match import GraphMatch

class GraphIndex:
    def __init__(self, metadata_attributes, metadata_index_by=None):
        self.images = {}
        self.pairs_index = {}

        self.unique_image_pairs = []
        self.inverse_unique_image_pairs = {}

        self.metadata_attributes = metadata_attributes
        # Inv. Indexes by Attribute
        self.metadata_index_by = metadata_index_by
        self.metadata_indices = {}

    def optimize_for_search(self):
        for entry_id in self.pairs_index:
            self.pairs_index[entry_id].optimize_for_search()

        for image_id in self.images:
            self.images[image_id].spatial_graph.update_index()

        self.inverse_unique_image_pairs = None

    def optimize_for_storage(self):
        for image_id in self.images:
            self.images[image_id].optimize_for_storage()

        self.inverse_unique_image_pairs = None


    def get_image_attribute(self, image_id, attribute):
        if image_id in self.images:
            att_pos = self.metadata_attributes.index(attribute)

            values = []
            for loc_idx, record in enumerate(self.images[image_id].source_metadata):
                loc_value = self.images[image_id].source_ids[loc_idx]
                values.append((loc_value, self.images[image_id].source_metadata[loc_idx][att_pos]))

            return values
        else:
            return None

    def get_images_by_attribute_value(self, attribute, value):
        matched_images = []
        if not attribute in self.metadata_indices:
            # requires linear scan (slow) ...
            att_pos = self.metadata_attributes.index(attribute)
            for image_id in self.images:
                for loc_idx, record in enumerate(self.images[image_id].source_metadata):
                    if record[att_pos] == value:
                        match_info = self.images[image_id].source_ids[loc_idx], image_id, self.images[image_id]
                        matched_images.append(match_info)
        else:
            # use pre-built index (fast) ...
            if value in self.metadata_indices[attribute]:
                # matches found .... get references ...
                for loc_value, image_id in self.metadata_indices[attribute][value]:
                    matched_images.append((loc_value, image_id, self.images[image_id]))

        return matched_images

    def add_image_with_pairs(self, file_path, source_ids, graph, pairs, source_metadata):
        next_id = len(self.images)
        self.images[next_id] = IndexImage(next_id, file_path, source_ids, graph, source_metadata)

        # if parallel indices need to be created by metadata attributes ...
        if self.metadata_index_by is not None:
            # add to corresponding indices ...
            self.index_image_metadata(next_id, source_ids, source_metadata)

        for raw_pair in pairs:
            self.add_pair(next_id, raw_pair)

    def index_image_metadata(self, image_id, source_ids, source_metadata):
        # ... for each attribute ....
        for attribute in self.metadata_index_by:
            # ... attribute position in metadata record
            att_pos = self.metadata_attributes.index(attribute)

            # ... if the index is empty ....
            if not attribute in self.metadata_indices:
                self.metadata_indices[attribute] = {}

            # ... for each location-based record for this image ...
            for location_idx, record in enumerate(source_metadata):
                # ... location ...
                loc_value = source_ids[location_idx]
                # ... current attribute value ....
                att_value = record[att_pos]

                # ... check if new value ....
                if not att_value in self.metadata_indices[attribute]:
                    self.metadata_indices[attribute][att_value] = []

                self.metadata_indices[attribute][att_value].append((loc_value, image_id))

    def add_pair(self, img_id, raw_pair):
        assert isinstance(raw_pair, GraphRawPair)

        # check if entry exists for pair ...
        entry_id = GraphIndexEntry.GenerateID(raw_pair.label_1, raw_pair.label_2)
        if not entry_id in self.pairs_index:
            # add entry to the index ...
            self.pairs_index[entry_id] = GraphIndexEntry(raw_pair.label_1, raw_pair.label_2)

        # finally add pair to the posting list ...
        # ... generate unique pair id by adding image id to local pair id ...
        global_pair_id = str(img_id) + "|" + raw_pair.pair_id
        if not global_pair_id in self.inverse_unique_image_pairs:
            # create numeric id for this unique global ...
            new_pair_id = len(self.unique_image_pairs)
            # ... inverse mapping ...
            self.inverse_unique_image_pairs[global_pair_id] = new_pair_id
            # ... save original LOCAL pair id ...
            self.unique_image_pairs.append(raw_pair.pair_id)
        else:
            new_pair_id = self.inverse_unique_image_pairs[global_pair_id]

        # replace string id with numerical id ...
        raw_pair.pair_id = new_pair_id

        self.pairs_index[entry_id].add_instance(img_id, raw_pair)

    def find_all_label_entries(self, label):
        matching_ids = []
        for entry_id in self.pairs_index:
            if self.pairs_index[entry_id].label_1 == label or self.pairs_index[entry_id].label_2 == label:
                matching_ids.append(entry_id)

        return matching_ids

    def get_raw_pair_partial_matches(self, raw_pairs, verbose=False):
        partial_matches = {}
        count_partial_matches = 0

        for raw_pair in raw_pairs:
            # assert isinstance(raw_pair, KeyFrameGraphRawPair)

            if raw_pair.label_1 == SpatialCCGraph.WildcardLabel and raw_pair.label_2 == SpatialCCGraph.WildcardLabel:
                if verbose:
                    print("-> Ignoring wildcard to wildcard pair")
                continue

            if (IndexParameters.Match_Wildcards and
                (raw_pair.label_1 == SpatialCCGraph.WildcardLabel or raw_pair.label_2 == SpatialCCGraph.WildcardLabel)):
                # single-wildcard pair. Do the matching .... (if wildcard matching is active)
                if raw_pair.label_1 != SpatialCCGraph.WildcardLabel:
                    concrete_label = raw_pair.label_1
                    concrete_first = True
                else:
                    concrete_label = raw_pair.label_2
                    concrete_first = False

                # print(raw_pair)

                # get all entries that match with the wildcard ...
                all_entry_id = self.find_all_label_entries(concrete_label)

                # about the order assumption ...
                # ... some orders require labels to be "flipped" (direction must be inverted!)
                # ... if concrete is "swapped", then direction/vector should be swapped too!
                inv_raw_pair = GraphRawPair.InvertPair(raw_pair)

                for entry_id in all_entry_id:
                    index_entry = self.pairs_index[entry_id]
                    if ((concrete_first and index_entry.label_1 == concrete_label) or
                        (not concrete_first and index_entry.label_2 == concrete_label)):
                        search_pair = raw_pair
                    else:
                        search_pair = inv_raw_pair

                    matches_scores, matches_data, matching_direction = index_entry.match(search_pair)
                    count_partial_matches += matches_scores.shape[0]

                    # if does not exist, create temporal list for this pair
                    if not raw_pair.pair_id in partial_matches:
                        partial_matches[raw_pair.pair_id] = []

                    # add to temporal list for this pair (note this is a list of Record-arrays)
                    partial_matches[raw_pair.pair_id].append((matches_scores, matches_data, matching_direction))

                    """
                    if entry_id == "2|R":
                        print(index_entry.optimized_meta_data)
                        print(partial_matches[raw_pair.pair_id])
                        raise Exception("STOP")
                    """
            else:
                # regular pair...
                # check if pair exist in index ...
                entry_id = GraphIndexEntry.GenerateID(raw_pair.label_1, raw_pair.label_2)

                # if exist ...
                if entry_id in self.pairs_index:
                    # score matches ...
                    matches_scores, matches_data, matching_direction = self.pairs_index[entry_id].match(raw_pair)
                    count_partial_matches += matches_scores.shape[0]

                    # if does not exist, create temporal list for this pair
                    if not raw_pair.pair_id in partial_matches:
                        partial_matches[raw_pair.pair_id] = []

                    # add to temporal list for this pair (note this is a list of Record-arrays)
                    partial_matches[raw_pair.pair_id].append((matches_scores, matches_data, matching_direction))

        return partial_matches, count_partial_matches

    def aggregate_raw_pair_partial_matches(self, partial_matches):
        # ... first aggregate scores by unique (query_pair_id, candidate_pair_id) ....
        count_merged_matches = 0
        all_candidates = []
        for query_pair_id in partial_matches:
            pair_matches = partial_matches[query_pair_id]

            merged_matches = GraphIndexEntry.merge_matches_by_pair(pair_matches)
            all_candidates.append((query_pair_id, merged_matches))

            count_merged_matches += len(merged_matches)

        # ... Then, divide per image ...
        per_image_candidates = {}
        for query_pair_id, q_pair_candidates in all_candidates:
            for img_id, unique_pair_id, score, direction in q_pair_candidates:
                if not img_id in per_image_candidates:
                    per_image_candidates[img_id] = []

                per_image_candidates[img_id].append((query_pair_id, unique_pair_id, score, direction))

        return per_image_candidates, count_merged_matches

    def compute_image_pre_ranking(self, query_graph, per_image_candidates):
        initial_ranking = []
        for img_id in per_image_candidates:
            # optimistically, assume that the highest scoring match for each query pair with a match will be kept
            # top_query_pairs = {}
            top_qc_matches = {}
            for query_pair_id, unique_pair_id, score, direction in per_image_candidates[img_id]:
                """
                # Rank 1: greedy best match per query pair
                if not query_pair_id in top_query_pairs:
                    # keep first score seen for a given query pair ..
                    top_query_pairs[query_pair_id] = score
                else:
                    # check if better match ...
                    if score > top_query_pairs[query_pair_id]:
                        # keep new best match ...
                        top_query_pairs[query_pair_id] = score
                """

                # Rank v-2: greedy best 1-to-1 match between query and candidate pairs ...
                if not query_pair_id in top_qc_matches:
                    top_qc_matches[query_pair_id] = {}

                # note: here we assume that pairs have been already aggregated, each
                top_qc_matches[query_pair_id][unique_pair_id] = score


            # Rank 2: greedily and optimistically choose best 1-to-1 pair matches between query and candidate
            # first, list all scored matches
            top_pair_matches = []
            for query_pair_id in top_qc_matches:
                for unique_pair_id in top_qc_matches[query_pair_id]:
                    match_info = (top_qc_matches[query_pair_id][unique_pair_id], query_pair_id, unique_pair_id)
                    top_pair_matches.append(match_info)

            # ... sort by descending score ...
            top_pair_matches = sorted(top_pair_matches, reverse=True, key=lambda x: x[0])

            # ... compute the score for 1-to-1 matches ...
            matched_query_pairs = {}
            matched_image_pairs = {}
            total_weighted_pair_score = 0.0
            for pair_score, query_pair_id, unique_pair_id in top_pair_matches:
                if not query_pair_id in matched_query_pairs and not unique_pair_id in matched_image_pairs:
                    # mark these elements as matched
                    matched_query_pairs[query_pair_id] = unique_pair_id
                    matched_image_pairs[unique_pair_id] = query_pair_id

                    # add the score ...
                    total_weighted_pair_score += pair_score

            # Note: the recall is not being divided by total query pairs
            #       since this is a common denominator for all images

            # (Rank v1)
            # add up all the top scores for each query pair for this image
            # max_edge_recall = sum([top_query_pairs[query_pair_id] for query_pair_id in top_query_pairs])

            # (Rank v2)
            max_edge_recall = total_weighted_pair_score
            max_edge_precision = max_edge_recall / self.images[img_id].spatial_graph.total_connections()

            score_info = (max_edge_recall, max_edge_precision)

            if IndexParameters.Layer1_ReturnRankings:
                # special case, keep extra information about the match
                # ... create a general match object for the approximate match
                img_graph = self.images[img_id].spatial_graph

                # Use greedily aligned pairs to create a match
                # ... note these pairs might be inconsistent, incompatible, etc.
                pair_alignments = []
                query_nodes = []
                image_nodes = []
                for query_pair_id in matched_query_pairs:
                    # get the pair ids
                    unique_pair_id = matched_query_pairs[query_pair_id]
                    local_pair_id = self.unique_image_pairs[unique_pair_id]

                    # ... add pair alignment ...
                    pair_alignments.append((query_pair_id, local_pair_id))

                    # ... estimate aligned nodes (query and candidates)
                    # ... Note: these aren't truly aligned as they would on the full aligned match
                    q_n1_str, q_n2_str = query_pair_id.split("|")
                    c_n1_str, c_n2_str = local_pair_id.split("|")

                    query_nodes.append(query_graph.vertices_by_id[q_n1_str])
                    query_nodes.append(query_graph.vertices_by_id[q_n2_str])

                    image_nodes.append(img_graph.vertices_by_id[c_n1_str])
                    image_nodes.append(img_graph.vertices_by_id[c_n2_str])

                approx_match = GraphMatch(query_graph, img_graph)
                approx_match.pair_alignments = pair_alignments
                approx_match.query_nodes = set(query_nodes)
                approx_match.candidate_nodes = set(image_nodes)

                # keep the current score ...
                # TODO: precision will be re-computed after match grouping. This should be avoided
                approx_match.total_score = max_edge_recall / query_graph.total_connections()

                # image id + approximated match
                match_data = (img_id, approx_match)
            else:
                # minimal information when all pairs are kept
                match_data = img_id

            initial_ranking.append((score_info, match_data))

        # finally, sort by descending score ...
        initial_ranking = sorted(initial_ranking, key=lambda x: x[0], reverse=True)

        # filter
        if 0 < IndexParameters.Top_Layer1_candidates < len(initial_ranking):
            # more candidates than requested. Keep the top K only!
            cut_position = IndexParameters.Top_Layer1_candidates - 1
            cut_scores = initial_ranking[cut_position][0]

            # find the first element with lower score than the element at the target position.
            while cut_position + 1 < len(initial_ranking) and initial_ranking[cut_position][0] >= cut_scores:
                cut_position += 1

            # cut all elements with lower score than the Top K-th element
            initial_ranking = initial_ranking[:cut_position]

        return initial_ranking

    def img_matches_from_img_ranking(self, query_graph, initial_ranking, per_image_candidates):
        final_candidates = {}
        # ... for each image at the top of initial rank ...
        for scores, match_data in initial_ranking:
            if IndexParameters.Layer1_ReturnRankings:
                # assume ranking contains an approximated match per candidate
                # add the single approximated match provided
                img_id, approx_match = match_data
                final_candidates[img_id] = [approx_match]
            else:
                # assume ranking only contains general information
                # ... create the pair-based matches
                img_id = match_data
                final_candidates[img_id] = []

                # for each matching pair ...
                for query_pair_id, unique_pair_id, score, direction in per_image_candidates[img_id]:
                    # creating candidate structure
                    local_pair_id = self.unique_image_pairs[unique_pair_id]
                    img_graph = self.images[img_id].spatial_graph
                    candidate = GraphMatch.FromPair(query_graph, query_pair_id, img_graph, local_pair_id,
                                                    score, direction)

                    # ... adding ...
                    final_candidates[img_id].append(candidate)

        return final_candidates

    def match_raw_pairs(self, query_graph, raw_pairs, verbose=False):
        query_stats = {}

        raw_list_start = time.time()

        # for each raw pair in the query ... find all raw pairwise matches
        partial_matches, count_partial_matches = self.get_raw_pair_partial_matches(raw_pairs, verbose)

        raw_list_end = time.time()
        query_stats["time_layer1_list_raw"] = raw_list_end - raw_list_start
        # print(count_partial_matches)
        # print("Time Step 1: " + str(raw_list_end - raw_list_start))

        merge_start = time.time()

        # now, merge raw pairwise matches by unique <query-pair-id, candidate-pair-id>
        per_image_candidates, count_merged_matches = self.aggregate_raw_pair_partial_matches(partial_matches)

        query_stats["count_layer1_initial_images"] = len(per_image_candidates)

        merge_end = time.time()
        query_stats["time_layer1_list_aggregate"] = merge_end - merge_start
        # print("Time Step 2: " + str(merge_end - merge_start))

        # compute pre-ranking scores per image
        pre_scoring_time = time.time()
        initial_ranking = self.compute_image_pre_ranking(query_graph, per_image_candidates)

        post_scoring_time = time.time()
        query_stats["time_layer1_pre_ranking"] = post_scoring_time - pre_scoring_time

        convert_start = time.time()

        # convert the ID's to actual references and pair nodes, and split by unique image ids ...
        final_candidates = self.img_matches_from_img_ranking(query_graph, initial_ranking, per_image_candidates)

        # mark as unused (garbage)
        per_image_candidates = None

        convert_end = time.time()

        # print("Time Step 3: " + str(convert_end - convert_start))
        query_stats["time_layer1_list_merge"] = convert_end - convert_start

        query_stats["count_layer1_query_pairs"] = len(raw_pairs)
        query_stats["count_layer1_raw_matches"] = count_partial_matches
        query_stats["count_layer1_aggregated_matches"] = count_merged_matches
        query_stats["count_layer1_images_matched"] = len(final_candidates)

        if verbose:
            print("-> Total Query Pairs: {0:d}".format(len(raw_pairs)), flush=True)
            print("-> Total Raw Matches: {0:d}".format(count_partial_matches), flush=True)
            print("-> Total Merged Matches: {0:d}".format(count_merged_matches), flush=True)
            print("-> Total Images Initially Matched: {0:d}".format(query_stats["count_layer1_initial_images"]))
            print("-> Total Images Finally Matched: {0:d}".format(len(final_candidates)), flush=True)

        return final_candidates, query_stats

    def total_pairs(self):
        return len(self.pairs_index)

    def total_instances(self):
        total = 0
        for entry_id in self.pairs_index:
            if self.pairs_index[entry_id].optimized_scores is not None:
                total += self.pairs_index[entry_id].optimized_scores.shape[0]
            else:
                total += self.pairs_index[entry_id].total_instances()

        return total

    def total_images(self):
        return len(self.images)

    def total_image_pairs(self):
        total = 0
        for image_id in self.images:
            total += self.images[image_id].spatial_graph.total_connections()

        return total

    def total_source_ids(self):
        return sum([len(self.images[image_id].source_ids) for image_id in self.images])

    def save_to_file(self, output_filename):
        with open(output_filename, "wb") as out_file:
            print("... Saving Index Images ...", flush=True)

            # save metadata attributes
            pickle.dump(self.metadata_attributes, out_file, pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.metadata_index_by, out_file, pickle.HIGHEST_PROTOCOL)
            pickle.dump(self.metadata_indices, out_file, pickle.HIGHEST_PROTOCOL)

            pickle.dump(self.unique_image_pairs, out_file, pickle.HIGHEST_PROTOCOL)

            # ... images ....
            pickle.dump(len(self.images), out_file, pickle.HIGHEST_PROTOCOL)
            for idx, image_id in enumerate(self.images):
                print("{0:d} / {1:d}".format(idx + 1, len(self.images)), end="\r", flush=True)
                pickle.dump((image_id, self.images[image_id]), out_file, pickle.HIGHEST_PROTOCOL)

                if idx % 10000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

            print("... Saving Index Posting lists ...", flush=True)
            pickle.dump(len(self.pairs_index), out_file, pickle.HIGHEST_PROTOCOL)
            for idx, pair_id in enumerate(self.pairs_index):
                print("{0:d} / {1:d}".format(idx + 1, len(self.pairs_index)), end="\r", flush=True)
                pickle.dump((pair_id, self.pairs_index[pair_id]), out_file, pickle.HIGHEST_PROTOCOL)

                if idx % 1000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

    @staticmethod
    def LoadFromFile(input_filename):

        with open(input_filename, "rb") as input_file:
            print("\n... Loading Index Images ...")

            # load metadata attributes
            metadata_attributes = pickle.load(input_file)
            metadata_index_by = pickle.load(input_file)
            metadata_indices = pickle.load(input_file)
            # ... unique pairs ...
            unique_image_pairs = pickle.load(input_file)

            graph_index = GraphIndex(metadata_attributes, metadata_index_by)
            graph_index.metadata_indices = metadata_indices

            graph_index.unique_image_pairs = unique_image_pairs

            n_images = pickle.load(input_file)
            for idx in range(n_images):
                print("{0:d} / {1:d}".format(idx + 1, n_images), end="\r")
                image_id, image_info = pickle.load(input_file)
                graph_index.images[image_id] = image_info

                if idx % 10000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

            print("\n... Loading Index Posting lists ...")
            n_postings = pickle.load(input_file)
            for idx in range(n_postings):
                print("{0:d} / {1:d}".format(idx + 1, n_postings), end="\r")
                pair_id, posting_list = pickle.load(input_file)
                graph_index.pairs_index[pair_id] = posting_list

                if idx % 1000 == 0:
                    # force memory release
                    gc.collect()

            # force memory release
            gc.collect()

        print("\n... Loading complete ...")

        return graph_index

