
import math
import os
import pickle
import time
import ctypes
from concurrent.futures import ProcessPoolExecutor

import cv2
import numpy as np
import scipy.ndimage.measurements as sci_mes

# general common tools
from AM_CommonTools.data.connected_component import ConnectedComponent
from AM_CommonTools.util.time_helper import TimeHelper
from AM_CommonTools.util.geometry_helper import GeometryHelper

# from current API ...
from TangentV.data.spatial_cc_graph import SpatialCCGraph
from TangentV.indexing.graph_raw_pair import GraphRawPair

# Interfaces with the DPRL Symbol Recognizer
from recognizer.data.feature_dataset import FeatureDataset
from recognizer.evaluation.eval_ops import EvalOps
from recognizer.features.feature_extractor import FeatureExtractor
from recognizer.preprocessing.inkml_loader import INKMLLoader
from recognizer.preprocessing.trace_smoother import TraceSmoother
from recognizer.recognition.symbol_classifier import SymbolClassifier
from recognizer.symbols.math_symbol import MathSymbol
from recognizer.symbols.math_symbol_trace import MathSymbolTrace
from recognizer.training.trainer import ClassifierTrainer

class RecognitionHelper:
    tangentv_lib = ctypes.CDLL('./tangentv_lib.so')

    @staticmethod
    def set_dataset_mapping(training_set, label_mapping=None, class_mapping=None):
        map_res = FeatureDataset.create_update_label_mapping(training_set.labels, label_mapping, class_mapping)
        mapped_labels, label_mapping, class_mapping = map_res

        training_set.labels = mapped_labels
        training_set.label_mapping = label_mapping
        training_set.class_mapping = class_mapping

    @staticmethod
    def get_contour_symbol(symbol_image, sym_id=0, label=None):
        # get contours from image ...
        symbol_image = cv2.copyMakeBorder(symbol_image, 2, 2, 2, 2, cv2.BORDER_CONSTANT, value=(0,))
        im2, contours, hierarchy = cv2.findContours(symbol_image, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

        # generate new traces from contours ...
        new_traces = []
        for contour_idx, contour in enumerate(contours):
            trace_points = contour[:, 0, :].tolist()
            trace_points.append((contour[0, 0, 0], contour[0, 0, 1]))

            contour_trace = MathSymbolTrace(contour_idx, trace_points)
            contour_trace.sharp_points = TraceSmoother.getSharpPoints(contour_trace)
            new_traces.append(contour_trace)

        # create the new symbol....
        new_symbol = MathSymbol(sym_id, new_traces, label)
        new_symbol.normalize()

        return new_symbol

    @staticmethod
    def split_training_set_by_classes(training_set, class_list):
        # use the list of classes to create two partitions of the dataset:
        # 1) samples only from the classes in the given list
        # 2) samples from the remaining classes

        class_samples = []
        class_labels = []
        non_class_samples = []
        non_class_labels = []

        for idx, label in enumerate(training_set.labels):
            if label in class_list:
                class_samples.append(training_set.data[idx, :])
                class_labels.append(label)
            else:
                non_class_samples.append(training_set.data[idx, :])
                non_class_labels.append(label)

        class_samples = np.array(class_samples)
        non_class_samples = np.array(non_class_samples)

        # create the corresponding datasets
        class_dataset = FeatureDataset(class_samples, class_labels, None, None)
        non_class_dataset = FeatureDataset(non_class_samples, non_class_labels, None, None)

        # copy normalizer as well ...
        class_dataset.normalizer = training_set.normalizer
        non_class_dataset.normalizer = training_set.normalizer

        return class_dataset, non_class_dataset

    @staticmethod
    def visualize_classifications(binary_image, accepted_labels, rejected_cc, out_prefix):
        # first, create basic template ...
        h, w = binary_image.shape
        rgb_image = np.zeros((h, w, 3), np.uint8)

        rgb_image[:, :, 0] = binary_image.copy()
        rgb_image[:, :, 1] = binary_image.copy()
        rgb_image[:, :, 2] = binary_image.copy()

        # ... rejection image ...
        reject_image = rgb_image.copy()
        for cc in rejected_cc:
            cc_cut = reject_image[cc.min_y:cc.max_y + 1, cc.min_x:cc.max_x +1, :]
            cc_cut[cc.img > 0, 0] = 0
            cc_cut[cc.img > 0, 1] = 0
            cc_cut[cc.img > 0, 2] = 255

        cv2.imwrite(out_prefix + "_rejected_cc.png", reject_image)

        # ... all other labels ...
        for label in accepted_labels:
            label_image = rgb_image.copy()
            for cc in accepted_labels[label]:
                cc_cut = label_image[cc.min_y:cc.max_y + 1, cc.min_x:cc.max_x +1, :]
                cc_cut[cc.img > 0, 0] = 0
                cc_cut[cc.img > 0, 1] = 255
                cc_cut[cc.img > 0, 2] = 0

            cv2.imwrite(out_prefix + "_label_" + label + "_cc.png", label_image)

    @staticmethod
    def visualize_node_top_class(binary_image, valid_ccs, valid_cc_classes, rejected_cc, prefix):
        ccs_per_label = {}
        for idx, cc in enumerate(valid_ccs):
            top_class, top_conf = valid_cc_classes[idx][0]

            if not top_class in ccs_per_label:
                ccs_per_label[top_class] = []

            ccs_per_label[top_class].append(cc)

        RecognitionHelper.visualize_classifications(binary_image, ccs_per_label, rejected_cc, prefix)

    @staticmethod
    def generate_merged_ccs_from_graph(vertices, connections, full_merge=True):
        # generate fake CCs for potential merges ...
        merged_ccs = []
        merged_ccs_index = {}
        for row, (idx1, idx2) in enumerate(connections):
            vertex_1 = vertices[idx1]
            vertex_2 = vertices[idx2]

            if full_merge:
                # create full merged of CC
                tempo_cc = ConnectedComponent.Merge([vertex_1, vertex_2])
            else:
                # merged box only ...
                merge_box = GeometryHelper.boxesBoudingBox([vertex_1.getBoundingBox(), vertex_2.getBoundingBox()])
                (min_x, max_x), (min_y, max_y) = merge_box

                # create fake CC (only the box)
                tempo_cc = ConnectedComponent(0, min_x, max_x, min_y, max_y, 0, None)

            # add ...
            merged_ccs.append(tempo_cc)
            merged_ccs_index[tempo_cc.strID()] = row

        return merged_ccs, merged_ccs_index

    @staticmethod
    def filter_cc_by_relative_scale(cc_list, median_h, median_w, min_scale, max_scale):
        if min_scale is None or max_scale is None:
            # no filter applied, simply return the original set of CC
            return cc_list, []

        valid_cc, rejected_cc = [], []
        for cc in cc_list:
            scale_h = cc.getHeight() / median_h
            scale_w = cc.getWidth() / median_w

            # check
            # 1) The maximum of the two scale factors must be greater than the minimum scale and
            # 2) Neither scale factor (horizontal or vertical) can be greater than the maximum scale ...
            # This prevents classifying components that are very long or very width, or that are too small.
            if min_scale <= max(scale_h, scale_w) and (scale_h <= max_scale and scale_w <= max_scale):
                # acceptable scaling ... process
                valid_cc.append(cc)
            else:
                rejected_cc.append(cc)

        return valid_cc, rejected_cc

    @staticmethod
    def get_cc_images_features(image_list, config, scaler):
        all_features = []
        feature_extractor = FeatureExtractor(config)
        for idx, image in enumerate(image_list):
            if idx % 10 == 0:
                print("Getting features from image: {0:d} out of {1:d}".format(idx + 1, len(image_list)), end="\r")
            contour_symbol = RecognitionHelper.get_contour_symbol(image)

            features = feature_extractor.getFeatures(contour_symbol)
            all_features.append(features)

        all_features = np.array(all_features)
        if scaler is not None:
            all_features = scaler.transform(all_features)

        return all_features

    @staticmethod
    def classify_ccs(classifier, cc_list, top_n):
        # first, get the contour symbols and extract their features ...
        image_list = [cc.img for cc in cc_list]
        all_features = RecognitionHelper.get_cc_images_features(image_list, classifier.config, classifier.scaler)

        # classify with confidence ...
        predicted = classifier.trained_classifier.predict_proba(all_features)

        n_classes = predicted.shape[1]
        if top_n is None or top_n > n_classes:
            top_n = n_classes

        classes = classifier.trained_classifier.classes_
        all_confidences = []
        for idx in range(len(cc_list)):
            current_confidences = predicted[idx, :].tolist()

            scores = sorted([(current_confidences[k], k) for k in range(n_classes)], reverse=True)

            confidences = [(classifier.classes_list[classes[scores[k][1]]], scores[k][0]) for k in range(top_n)]
            all_confidences.append(confidences)

        return all_confidences

    @staticmethod
    def test_merge_ccs(initial_classes, neighbor_graph, classifier, top_n):
        # get the graph edges ...
        connections = neighbor_graph.get_connections()

        # get the merged  CCs ....
        merged_ccs, merged_ccs_index = RecognitionHelper.generate_merged_ccs_from_graph(neighbor_graph.vertices,
                                                                                        connections, True)
        # classify merged CCs ....
        merged_cc_classes = RecognitionHelper.classify_ccs(classifier, merged_ccs, top_n)

        accepted_merges = []
        for idx, pair_classes in enumerate(merged_cc_classes):
            vertex_idx_1, vertex_idx_2 = connections[idx]

            top_class_1, top_conf_1 = initial_classes[vertex_idx_1][0]
            top_class_2, top_conf_2 = initial_classes[vertex_idx_2][0]

            # <DEBUG: Check a merge ... >
            """
            check_ids = ['465-509-107-110-178', '465-509-120-123-178']
            if (neighbor_graph.vertices[vertex_idx_1].strID() in check_ids and
                neighbor_graph.vertices[vertex_idx_2].strID() in check_ids):
                print("\n.....")
                print(initial_classes[vertex_idx_1])
                print(initial_classes[vertex_idx_2])
                print(pair_classes)
                print(sorted(classifier.classes_list))
                print("\n")
                raise Exception("DEBUG")
            """
            # </DEBUG>

            average_conf = (top_conf_1 + top_conf_2) / 2.0

            top_merged_class, top_merged_conf = pair_classes[0]
            if top_merged_class != "JUNK" and (top_merged_conf >= average_conf):
                # print(((top_class_1, top_conf_1), (top_class_2, top_conf_2), (top_merged_class,top_merged_conf)))
                accepted_merges.append((top_merged_conf, vertex_idx_1, vertex_idx_2, merged_cc_classes[idx]))

        accepted_merges = sorted(accepted_merges, reverse=True)
        final_merges = []
        accepted_idxs = []
        for top_conf, v_idx_1, v_idx_2, classes in accepted_merges:
            if (v_idx_1 not in accepted_idxs) and (v_idx_2 not in accepted_idxs):
                # mark as accepted ...
                accepted_idxs.append(v_idx_1)
                accepted_idxs.append(v_idx_2)

                final_merges.append((v_idx_1, v_idx_2, classes))

        return final_merges

    @staticmethod
    def filter_neighbors_by_merge_size(neighbor_graph, median_w, median_h, min_scale=0.333, max_scale=3.0):
        # get the graph edges ...
        # TODO: this might fail if the graph contains loops (assumed not to have loops when this is called)
        connections = neighbor_graph.get_connections()

        fake_ccs, fake_ccs_index = RecognitionHelper.generate_merged_ccs_from_graph(neighbor_graph.vertices, connections, False)

        # filter fake CCs ...
        accepted, rejected = RecognitionHelper.filter_cc_by_relative_scale(fake_ccs, median_h, median_w, min_scale, max_scale)

        # remove rejected links from graph...
        for fake_cc in rejected:
            # get the node indices
            row = fake_ccs_index[fake_cc.strID()]
            idx1, idx2 = connections[row]

            # remove connections ...
            neighbor_graph.connections[idx1][idx2] = False
            neighbor_graph.connections[idx2][idx1] = False

    @staticmethod
    def CCsFromBinary(binary_image, min_size=None):
        labels, count_labels = sci_mes.label(binary_image)
        h, w = binary_image.shape

        # no CC?
        if count_labels == 0:
            return []

        # input
        labels_p = labels.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))

        # output arrays ...
        out_mins_y = np.zeros(count_labels, dtype=np.int32)
        out_maxs_y = np.zeros(count_labels, dtype=np.int32)
        out_mins_x = np.zeros(count_labels, dtype=np.int32)
        out_maxs_x = np.zeros(count_labels, dtype=np.int32)
        out_counts = np.zeros(count_labels, dtype=np.int32)

        # output pointers ....
        out_mins_y_p = out_mins_y.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))
        out_maxs_y_p = out_maxs_y.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))
        out_mins_x_p = out_mins_x.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))
        out_maxs_x_p = out_maxs_x.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))
        out_counts_p = out_counts.ctypes.data_as(ctypes.POINTER(ctypes.c_int32))

        # call C function that finds boundaries and pixel count for each CC on single pass.
        """
          int* labels, int width, int height, int count_labels,
          int* out_mins_y, int* out_maxs_y, int* out_mins_x, 
          int* out_maxs_x, int* out_counts
        """
        arg_types = [
            ctypes.POINTER(ctypes.c_int32), ctypes.c_int32, ctypes.c_int32, ctypes.c_int32,
            ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32),
            ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_int32)
        ]
        RecognitionHelper.tangentv_lib.CC_Boundaries.argtypes = arg_types
        RecognitionHelper.tangentv_lib.CC_Boundaries.restype = ctypes.c_int32
        RecognitionHelper.tangentv_lib.CC_Boundaries(labels_p, w, h, count_labels,
                                                     out_mins_y_p, out_maxs_y_p, out_mins_x_p,
                                                     out_maxs_x_p, out_counts_p)

        components = []
        for cc_id in range(count_labels):
            # check size ....
            if min_size is None or out_counts[cc_id] >= min_size:
                max_x = out_maxs_x[cc_id]
                max_y = out_maxs_y[cc_id]
                min_x = out_mins_x[cc_id]
                min_y = out_mins_y[cc_id]

                cc_img = (labels[min_y:max_y + 1, min_x:max_x + 1] == cc_id + 1).astype(np.uint8) * 255

                component = ConnectedComponent(cc_id, min_x, max_x, min_y, max_y, out_counts[cc_id], cc_img)
                components.append(component)

        return components

    @staticmethod
    def CCsFromLabeledBoxes(labeled_boxes, avg_pixel_height=60, density=0.5):
        # first find the boundaries of the boxes
        all_x, all_y, all_w, all_h, cc_classes = zip(*labeled_boxes)

        # previously ...
        # cc_classes = [[(current_class, 1.0)] for current_class in cc_classes]
        new_cc_classes = []
        for current_classes in cc_classes:
            # each class brings a confidence which in most cases will be 1.0
            new_cc_classes.append([(label, current_classes[label]) for label in current_classes])

        cc_classes = new_cc_classes

        # work as numpy arrays ...
        all_min_x = np.array(all_x)
        all_min_y = np.array(all_y)
        all_w = np.array(all_w)
        all_h = np.array(all_h)
        all_max_x = all_min_x + all_w
        all_max_y = all_min_y + all_h

        # find global displacement
        bb_min_x = all_min_x.min()
        bb_min_y = all_min_y.min()

        # apply translation ...
        all_min_x -= bb_min_x
        all_max_x -= bb_min_x
        all_min_y -= bb_min_y
        all_max_y -= bb_min_y

        # find average symbol heights ...
        avg_h = all_h.mean()

        # get scaling factor ...
        scale = avg_pixel_height / avg_h

        # apply scale
        all_min_x *= scale
        all_max_x *= scale
        all_min_y *= scale
        all_max_y *= scale

        w = int(all_max_x.max()) + 1
        h = int(all_max_y.max()) + 1

        # now, generate the new CCs
        components = []
        for idx in range(len(labeled_boxes)):
            min_x = int(all_min_x[idx])
            min_y = int(all_min_y[idx])
            max_x = int(all_max_x[idx])
            max_y = int(all_max_y[idx])
            # use a size based on
            cc_size = int(((max_x - min_x + 1) * (max_y - min_y + 1)) * density)
            new_cc = ConnectedComponent(idx, min_x, max_x, min_y, max_y, cc_size, None)
            components.append(new_cc)
            # print(new_cc)

        return components, (w, h), cc_classes

    @staticmethod
    def classify_image_ccs(binary_image, classifier, knn, top_n=10, min_scale=0.333, max_scale=3.0, max_edge_dist=2.0,
                           render_prefix=None, verbose=False, wildcard_ccs=None):

        assert isinstance(classifier, SymbolClassifier)

        h, w = binary_image.shape

        # get the CC from the image ...
        current_cc = RecognitionHelper.CCsFromBinary(binary_image)
        if len(current_cc) == 0:
            return None, None, None

        # get size stats ...
        median_h, median_w = ConnectedComponent.MedianSize(current_cc)
        if verbose:
            print("-> Median CC Size: {0:.2f} x {1:.2f} (h x w)".format(median_h, median_w))

        median_distance = max(median_h, median_w)

        # create LOS graph, do not connect elements that are too far way ...
        graph = SpatialCCGraph(current_cc, w, h)
        graph.makeLOS(SpatialCCGraph.PolygonConvexHull, median_distance * max_edge_dist,
                      SpatialCCGraph.DistanceContours)
        if verbose:
            print("-> Initial Graph vertices: " + str(len(graph.vertices)))
            print("-> Initial Graph connections: " + str(graph.total_connections()))

        # filter CCs by size (remove too big or too small)...
        valid_cc, rejected_cc = RecognitionHelper.filter_cc_by_relative_scale(current_cc, median_h, median_w,
                                                                              min_scale, max_scale)

        # remove rejected CC from graph..
        for cc in rejected_cc:
            graph.delete(cc.strID())

        if verbose:
            print("-> Graph vertices after filter by size: " + str(len(graph.vertices)))
            print("-> Graph connections after filter by size: " + str(graph.total_connections()))

        # get initial classification ...
        start_time = time.time()
        valid_cc_classes = RecognitionHelper.classify_ccs(classifier, valid_cc, top_n)
        end_time = time.time()
        if verbose:
            print("-> Total classification time: " + TimeHelper.secondsToStr(end_time - start_time))

        # check for merging ...
        segment_graph = SpatialCCGraph(valid_cc, w, h)
        # segment_graph.makeLOS(SpatialCCGraph.PolygonConvexHull, median_distance * 0.5, SpatialCCGraph.DistanceContours)
        segment_graph.makeKNN(knn, SpatialCCGraph.DistanceContours)
        if verbose:
            print("-> Initial K-nn connections: " + str(segment_graph.total_connections()))

        # Disconnect wildcard CC's on segment to prevent them from being merged!
        if wildcard_ccs is not None and len(wildcard_ccs) > 0:
            for wildcard_cc in wildcard_ccs:
                # Warning: pixel count could mismatch here!
                segment_graph.disconnect_node(wildcard_cc.strID())

        # apply filter to remove pairs that would create elements that are too large ...
        RecognitionHelper.filter_neighbors_by_merge_size(segment_graph, median_w, median_h, min_scale, max_scale)
        if verbose:
            print("-> Total K-nn connections to test for merging CCs: " + str(segment_graph.total_connections()))

        # evaluate potential merges ...
        if segment_graph.total_connections() > 0:
            accepted_merges = RecognitionHelper.test_merge_ccs(valid_cc_classes, segment_graph, classifier, top_n)
            tempo_merged = [(min(v_1, v_2), max(v_1, v_2), cc_class) for v_1, v_2, cc_class in accepted_merges]
            accepted_merges = sorted(tempo_merged, key=lambda x:x[1], reverse=True)
        else:
            accepted_merges = []

        # sort merges by largest vertex idx from higher to lower.
        # In this way, indices of remaining pairs to merge will still be consistent with original ordering ...
        if verbose:
            print("-> Total accepted merges: " + str(len(accepted_merges)))
        # print(accepted_merges)

        for v_idx_1, v_idx_2, merged_cc_classes in accepted_merges:
            # assuming v_idx_1 < v_idx_2
            vertex_1 = valid_cc[v_idx_1]
            vertex_2 = valid_cc[v_idx_2]

            graph.merge_vertices(vertex_1.strID(), vertex_2.strID())

            # update classes ...
            del valid_cc_classes[v_idx_2]
            valid_cc_classes[v_idx_1] = merged_cc_classes

        if render_prefix is not None:
            RecognitionHelper.visualize_node_top_class(binary_image, graph.vertices, valid_cc_classes, rejected_cc,
                                                       render_prefix)

        if verbose:
            print("-> Final Graph connections: " + str(graph.total_connections()))

        return graph, valid_cc_classes, rejected_cc

    @staticmethod
    def LabeledBoxesToGraph(labeled_boxes, max_edge_distance, verbose=False):
        # check for empty pdf ...
        if len(labeled_boxes) == 0:
            return None, None

        # get the CC from PDF boxes ...
        current_cc, (w, h), cc_classes = RecognitionHelper.CCsFromLabeledBoxes(labeled_boxes)

        # get size stats ...
        median_h, median_w = ConnectedComponent.MedianSize(current_cc)
        if verbose:
            print("-> Median CC Size: {0:.2f} x {1:.2f} (h x w)".format(median_h, median_w))

        median_distance = max(median_h, median_w)

        # create LOS graph, do not connect elements that are too far way ...
        graph = SpatialCCGraph(current_cc, w, h)
        max_distance = median_distance * max_edge_distance
        graph.makeLOS(SpatialCCGraph.PolygonBoundingBox, max_distance, SpatialCCGraph.DistanceBoxes)

        if verbose:
            print("-> Graph vertices: " + str(len(graph.vertices)))
            print("-> Graph connections: " + str(graph.total_connections()))

        return graph, cc_classes

    @staticmethod
    def top_classes_by_min_confidence(class_list, confidences, min_class_confidence):
        # assume that list of classes are given with a confidence
        total_confidence = 0.0
        final_labels = []
        for idx, class_label in enumerate(class_list):
            class_conf = confidences[idx]
            if class_label != "JUNK":
                final_labels.append((class_label, class_conf))

            total_confidence += class_conf
            if total_confidence >= min_class_confidence:
                break

        return final_labels

    @staticmethod
    def get_index_pairs(graph, min_class_confidence, verbose=True):
        connections = graph.get_connections()

        # pre-built unit vector cache
        graph.buildUnitVector3DCache()

        # filter labels per CC to keep only the ones that will be used for pairs ...
        top_classes = []
        for node_idx, class_list in enumerate(graph.vertices_labels):
            confidences = graph.vertices_confidences[node_idx]
            top_classes.append(RecognitionHelper.top_classes_by_min_confidence(class_list, confidences,
                                                                               min_class_confidence))

        count_pair_combinations = []
        pair_combinations = []
        total_prod_max_weight = 0.0
        total_comb_max_weight = 0.0
        for idx, (v_idx_1, v_idx_2) in enumerate(connections):
            # for all combinations of labels ...
            v_1 = graph.vertices[v_idx_1]
            v_2 = graph.vertices[v_idx_2]

            # for product-based metrics, the maximum score for confidences is matching the pair with the
            # highest confidence using a query pair with perfect confidence (it can't do any better)
            # top_index_1 * 1.0  * top_index_2 * 1.0
            if len(top_classes[v_idx_1]) > 0 and len(top_classes[v_idx_2]) > 0:
                # there is a valid top class for both nodes ...
                top_class_v1, top_class_conf_1 = top_classes[v_idx_1][0]
                top_class_v2, top_class_conf_2 = top_classes[v_idx_2][0]

                total_prod_max_weight += top_class_conf_1 * top_class_conf_2

            # get pairs ...
            edge_info = GraphRawPair.FromGraphEdge(graph, v_1, top_classes[v_idx_1], v_2, top_classes[v_idx_2])
            edge_pairs, pairs_comb_weight = edge_info
            total_comb_max_weight += pairs_comb_weight
            # add to the complete list ...
            pair_combinations += edge_pairs

            count_pair_combinations.append(len(edge_pairs))

        if verbose:
            count_pair_combinations = np.array(count_pair_combinations)
            print("-> Total pairs of symbols in the graph: " + str(count_pair_combinations.size))
            if count_pair_combinations.size > 0:
                print("--> Average combinations per pair: " + str(count_pair_combinations.mean()))
                print("--> STDev combinations per pair: " + str(count_pair_combinations.std()))
                print("--> Total pair combinations: " + str(count_pair_combinations.sum()))

        return pair_combinations, (total_prod_max_weight, total_comb_max_weight)

    @staticmethod
    def show_class_list(label_mapping):
        label_mapping = sorted(label_mapping)

        classes_per_row = 10
        n_rows = int(math.ceil(len(label_mapping) / classes_per_row))

        for row in range(n_rows):
            print("\t".join(label_mapping[row * classes_per_row:(row + 1) * classes_per_row]))

        print("Total: {0:d} classes found".format(len(label_mapping)))

    @staticmethod
    def correct_training_data_classes(training_set, label_corrections, verbose=True):
        # get the original set of classes
        if verbose:
            print("Original set of classes")
            _, label_mapping, class_mapping = FeatureDataset.create_update_label_mapping(training_set.labels, None, None)

            RecognitionHelper.show_class_list(label_mapping)
            print("")

        for idx, label in enumerate(training_set.labels):
            if label in label_corrections:
                # re-map class ...
                training_set.labels[idx] = label_corrections[label]

        _, label_mapping, class_mapping = FeatureDataset.create_update_label_mapping(training_set.labels, None, None)

        if verbose:
            print("Corrected set of classes")
            RecognitionHelper.show_class_list(label_mapping)
            print("")

        return label_mapping

    @staticmethod
    def get_reference_training_metrics(training_set, testing_set, rec_config):
        # fix mapping
        RecognitionHelper.set_dataset_mapping(training_set)
        RecognitionHelper.set_dataset_mapping(testing_set, training_set.label_mapping, training_set.class_mapping)

        print("Full Training ...", flush=True)
        classifier = ClassifierTrainer.train_classifier(training_set, rec_config)

        # evaluate on lecture training set ...
        train_acc, train_avg, train_std, _ = EvalOps.classifier_per_class_accuracy(classifier, training_set)

        print("->Training Results")
        print("\tGlb. ACC\tPC. AVG\tPC. STD")
        print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(train_acc, train_avg, train_std))

        # evaluate on lecture testing set ...
        test_acc, test_avg, test_std, errors = EvalOps.classifier_per_class_accuracy(classifier, testing_set)
        print("->Testing Results")
        print("\tGlb. ACC\tPC. AVG\tPC. STD")
        print("\t{0:.4f}\t{1:.4f}\t{2:.4f}".format(test_acc, test_avg, test_std))

        return classifier

    @staticmethod
    def save_classified_symbols_in_chunks(tanv_config, all_graphs, ids_per_location, attributes, metadata_per_location):
        temporary_output_prefix = tanv_config.get("TEMPORARY_CLASSIFIED_SYMBOLS")
        temporary_chunk_size = tanv_config.get_int("TEMPORARY_CLASSIFIED_CHUNKS")

        n_chunks = int(math.ceil(len(all_graphs) / temporary_chunk_size))
        for idx in range(n_chunks):
            temporary_output_filename = temporary_output_prefix + "_" + str(idx) + ".dat"
            with open(temporary_output_filename, "wb") as out_file:
                if idx == 0:
                    # store this only once ... with the first chunk of data
                    pickle.dump(n_chunks, out_file, pickle.HIGHEST_PROTOCOL)
                    pickle.dump(ids_per_location, out_file, pickle.HIGHEST_PROTOCOL)
                    pickle.dump(attributes, out_file, pickle.HIGHEST_PROTOCOL)
                    pickle.dump(metadata_per_location, out_file, pickle.HIGHEST_PROTOCOL)

                chunk_start = idx * temporary_chunk_size
                chunk_end = (idx + 1) * temporary_chunk_size
                pickle.dump(all_graphs[chunk_start:chunk_end], out_file, pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def find_binary_class_threshold(positive_values, negative_values, min_recall):
        # find ideal threshold ...
        unique_values = sorted(list(set(positive_values + negative_values)))
        counts_per_value = {value: {"+": 0, "-": 0} for value in unique_values}

        # ... count values ...
        for value in positive_values:
            counts_per_value[value]["+"] += 1

        for value in negative_values:
            counts_per_value[value]["-"] += 1

        recall_per_value = []
        precision_per_value = []
        fmeasure_per_value = []
        pos_l = 0
        pos_r = len(positive_values)
        neg_l = 0
        neg_r = len(negative_values)

        change_value = None
        t_recall, t_precision, t_fmeasure = 0.0, 0.0, 0.0
        # accuracy_per_value.append(total_correct / (total_correct + total_incorrect))
        for idx, value in enumerate(unique_values):
            # as threshold increases, some positive samples will be erroneously classified as negative ...
            pos_l += counts_per_value[value]["+"]
            pos_r -= counts_per_value[value]["+"]

            # as threshold increases, some negative samples will be correctly classified as negative...
            neg_l += counts_per_value[value]["-"]
            neg_r -= counts_per_value[value]["-"]

            if pos_r > 0.0:
                recall = pos_r / len(positive_values)
            else:
                recall = 0.0

            if pos_r + neg_r > 0.0:
                precision = pos_r / (pos_r + neg_r)
            else:
                precision = 0.0

            if recall + precision > 0.0:
                fmeasure = (2.0 * precision * recall) / (recall + precision)
            else:
                fmeasure = 0.0

            recall_per_value.append(recall)
            precision_per_value.append(precision)
            fmeasure_per_value.append(fmeasure)

            if change_value is None:
                if recall < min_recall and idx > 0:
                    change_value = unique_values[idx - 1]
                    t_recall = recall_per_value[-2]
                    t_precision = precision_per_value[-2]
                    t_fmeasure = fmeasure_per_value[-2]
                elif precision >= recall:
                    change_value = value
                    t_recall = recall
                    t_precision = precision
                    t_fmeasure = fmeasure

        # plt.plot(unique_values, recall_per_value, 'r', unique_values, precision_per_value, 'b', unique_values, fmeasure_per_value, 'g')
        # plt.show()

        return change_value, (t_recall, t_precision, t_fmeasure)

    @staticmethod
    def find_confidence_threshold(classifier, train_dataset, junk_dataset, min_recall):
        assert isinstance(classifier, SymbolClassifier)

        predicted = classifier.trained_classifier.predict_proba(train_dataset.data)

        inv_mapping = {label: idx for idx, label in enumerate(classifier.trained_classifier.classes_)}

        class_confidences = []
        for idx in range(train_dataset.data.shape[0]):
            label = train_dataset.labels[idx, 0]
            class_conf = predicted[idx, inv_mapping[label]]
            class_confidences.append(class_conf)

        # now, repeat for junk class .... assume classifier has NOT been trained using junk class
        predicted = classifier.trained_classifier.predict_proba(junk_dataset.data)
        # pick highest confidence for each case ...
        junk_confidences = np.max(predicted, axis=1).tolist()

        # to show confidence histograms
        # f, axarr = plt.subplots(2, 1)
        # axarr[0].hist(class_confidences)
        # axarr[1].hist(junk_confidences)
        # plt.show()

        # find threshold ...
        threshold = RecognitionHelper.find_binary_class_threshold(class_confidences, junk_confidences, min_recall)

        return threshold

    @staticmethod
    def threshold_classifier_per_class_accuracy(symbol_classifier, feature_dataset, threshold, chunk_size=5000):
        # now, compute training evaluation metrics
        # ... predict for training set ....
        predicted = EvalOps.predict_topk_in_chunks(feature_dataset.data, symbol_classifier, 1, chunk_size)

        reject_predict = []
        for [(symbol_class, confidence)] in predicted:
            if confidence >= threshold:
                reject_predict.append(symbol_classifier.classes_dict[symbol_class])
            else:
                reject_predict.append(-1)
        # ... compute metrics...
        n_classes = feature_dataset.num_classes()

        return EvalOps.per_class_accuracy(reject_predict, feature_dataset.labels, n_classes)

    @staticmethod
    def threshold_classifier_junk_accuracy(symbol_classifier, junk_dataset, threshold, chunk_size=5000):
        # ... predict for junk data set ....
        predicted = EvalOps.predict_topk_in_chunks(junk_dataset.data, symbol_classifier, 1, chunk_size)

        n_samples = junk_dataset.data.shape[0]
        total_errors = 0
        for [(symbol_class, confidence)] in predicted:
            if confidence >= threshold:
                total_errors += 1

        return ((n_samples - total_errors) / n_samples) * 100.0

    @staticmethod
    def load_label_mapping(mapping_filename):
        with open(mapping_filename, "r", encoding="utf8") as out_file:
            mapping_lines = out_file.readlines()

        class_mapping = {}
        for line in mapping_lines:
            parts = line.strip().split("\t")
            if len(parts) == 2:
                class_idx, class_label = parts
                class_mapping[class_label] = int(class_idx)

        label_mapping = [None for i in range(len(class_mapping))]
        for class_label in class_mapping:
            label_mapping[class_mapping[class_label]] = class_label

        return label_mapping, class_mapping

    @staticmethod
    def load_extract_dir_img_features(feature_extractor, input_dir_info):
        # load all images in directory ...
        parent_dir, sub_dir_name = input_dir_info
        current_dir = parent_dir + "/" + sub_dir_name

        print("-> Extracting features from: " + current_dir)

        # extract features ...
        dir_features = []

        dir_image_names = os.listdir(current_dir)
        for img_name in dir_image_names:
            image_full_path = current_dir + "/" + img_name
            print("-> processing: " + image_full_path, flush=True)

            symbol_image = cv2.imread(image_full_path, cv2.IMREAD_GRAYSCALE)

            contour_symbol = RecognitionHelper.get_contour_symbol(symbol_image)
            features = feature_extractor.getFeatures(contour_symbol)

            dir_features.append(features)

        return sub_dir_name, dir_features

    @staticmethod
    def extract_features_from_img_dirs(root_dir, label_mapping, rec_config, normalizer, workers, verbose=True):
        sub_directories = [(root_dir, sub_dir_name) for sub_dir_name in os.listdir(root_dir)]

        # feature extractor ...
        feat = FeatureExtractor(rec_config)
        feat_refs = [feat] * len(sub_directories)

        if verbose:
            print("-> A total of {0:d} class(es) found ".format(len(sub_directories)))

        # use parallel threads to load data from each directory ...
        all_features = []
        raw_labels = []
        with ProcessPoolExecutor(max_workers=workers) as executor:
            for dir_name, dir_features in executor.map(RecognitionHelper.load_extract_dir_img_features,
                                                       feat_refs, sub_directories):
                count_features = len(dir_features)
                all_features += dir_features

                dir_label = label_mapping[int(dir_name)]
                raw_labels += [dir_label] * count_features

        data_features = np.array(all_features)
        # ... normalize features ...
        dataset = FeatureDataset(data_features, raw_labels, None, None)
        dataset.normalize_data(normalizer)

        return dataset

    @staticmethod
    def load_named_classifers_file(filename):
        # load from file ...
        with open(filename, "r", encoding="utf8") as input_file:
            all_classifier_lines = input_file.readlines()

        classifiers_by_name = {}
        for line in all_classifier_lines:
            parts = line.strip().split("\t")
            if len(parts) == 3:
                classifier_name, config_filename, classifier_filename = parts
                classifiers_by_name[classifier_name] = (config_filename, classifier_filename)

        return classifiers_by_name

    @staticmethod
    def load_named_classifers_from_config(config):
        tempo_classifiers = config.get("SYMBOL_CLASSIFIER_BY_NAME")

        # check if filename is provided or dict
        if isinstance(tempo_classifiers, str):
            classifiers_by_name = RecognitionHelper.load_named_classifers_file(tempo_classifiers)

            # update on config ...
            config.set("SYMBOL_CLASSIFIER_BY_NAME", classifiers_by_name)
        else:
            classifiers_by_name = tempo_classifiers

        return classifiers_by_name
