
import gc
import os
import time

import cv2
import numpy as np

from AM_CommonTools.util.time_helper import TimeHelper

from TangentV.indexing.recognition_helper import RecognitionHelper
from TangentV.indexing.graph_index import GraphIndex
from TangentV.indexing.index_parameters import IndexParameters
from TangentV.retrieval.graph_match import GraphMatch
from TangentV.retrieval.query_image import QueryImage
from TangentV.retrieval.match_helper import MatchHelper


class QueryInfo:
    TypeSingleImage = 0
    XMLNamespace = ''

    GroupModeNone = 0
    GroupModeImage = 1
    GroupModeSequence = 2
    GroupModeAttribute = 3

    PrecisionNone = 0
    PrecisionGlobalPairs = 1

    def __init__(self, query_id, query_set, query_type, group_mode, group_options=None):
        self.id = query_id
        self.query_set = query_set
        self.type = query_type
        self.group_mode = group_mode
        self.group_options = group_options

        self.query_images = []

    def preprocess_images(self, tanv_config, named_classifiers):
        # pre-process each query image ...
        for query_image in self.query_images:
            query_image.preprocess(tanv_config, named_classifiers)

    def execute(self, image_index):
        assert isinstance(image_index, GraphIndex)

        if self.type == QueryInfo.TypeSingleImage:
            all_raw_pairs = self.query_images[0].raw_pairs
            query_graph = self.query_images[0].graph

            return QueryInfo.execute_single_image(image_index, query_graph, all_raw_pairs,
                                                  self.group_mode, self.group_options)

    @staticmethod
    def execute_single_image(image_index, query_graph, all_raw_pairs, group_mode, group_options=None):
        query_stats = {}

        # First layer, hit the index and find all matching pairs for the set of raw query pairs ...
        start_layer_1 = time.time()
        gc.disable()
        pair_matches_per_img, layer_stats = image_index.match_raw_pairs(query_graph, all_raw_pairs, True)
        gc.enable()
        end_layer_1 = time.time()
        query_stats.update(layer_stats)
        query_stats["time_layer1_total"] = end_layer_1 - start_layer_1
        print("Total layer 1: " + str(end_layer_1 - start_layer_1))

        debug_img_ids = None
        debug_region = None
        # debug_img_ids = ['10|27550']
        # debug_region = (1264, 1626, 393, 495)

        # Second layer, grow full matche using the raw pairs found from the index.
        #   Then, connect compatible matches using pairs that align well spatially
        #   but where not matched by query pairs directly.
        start_layer_2 = time.time()
        gc.disable()

        img_matches = {}
        # prepare stat counters ...
        query_stats["time_layer2_grow_merge"] = 0.0
        query_stats["time_layer2_grow_connect"] = 0.0
        query_stats["time_layer2_match_filter"] = 0.0
        query_stats["time_layer2_join_disconnected"] = 0.0
        query_stats["count_layer2_initial_matches"] = 0
        query_stats["count_layer2_grown_matches"] = 0
        query_stats["count_layer2_connected_matches"] = 0
        query_stats["count_layer2_final_matches"] = 0
        query_stats["count_layer2_join_disconnected"] = 0
        for idx, img_id in enumerate(sorted(list(pair_matches_per_img.keys()))):
            if debug_img_ids is not None and not img_id in debug_img_ids:
                # skip key-frame ....
                continue

            candidate_graph = image_index.images[img_id].spatial_graph
            candidate_graph.buildUnitVector3DCache()
            matching_pairs = pair_matches_per_img[img_id]

            img_matches[img_id], img_stats = MatchHelper.align_img_matches(query_graph, candidate_graph, matching_pairs,
                                                                           debug_region)
            # Update global query stats ...
            for stat_name in img_stats:
                query_stats[stat_name] += img_stats[stat_name]
        gc.enable()
        end_layer_2 = time.time()
        query_stats["time_layer2_total"] = end_layer_2 - start_layer_2

        print("Total layer 2: " + str(end_layer_2 - start_layer_2))

        start_layer_3 = time.time()

        # based on selected group mode ..
        if group_mode == QueryInfo.GroupModeImage:
            # group per image ...
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_image(image_index, img_matches)
        elif group_mode == QueryInfo.GroupModeSequence:
            # group per sequence on attribute after grouping by parent attribute ...
            group_by_attribute = group_options["group_by"]
            sequence_by_attribute = group_options["sequence_by"]
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_sequence(image_index, img_matches,
                                                                                       group_by_attribute,
                                                                                       sequence_by_attribute)
        elif group_mode == QueryInfo.GroupModeAttribute:
            # group by set of attributes ....
            group_by_attributes = group_options["group_by"]
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_attributes(image_index, img_matches,
                                                                                         group_by_attributes)
        else:
            # by default, do not group ...
            final_grouped_matches, group_stats = MatchHelper.group_matches_by_none(image_index, img_matches)

        # Optional Precision-based second metric
        if IndexParameters.Layer3_PrecisionScore == QueryInfo.PrecisionGlobalPairs:
            # add a second score aside from main similarity metric, to break ties based on precision.
            final_grouped_matches = MatchHelper.add_precision_metric(final_grouped_matches,
                                                                     GraphMatch.getGlobalPairBasedPrecision)


        query_stats.update(group_stats)


        end_layer_3 = time.time()
        print("Total layer 3: " + str(end_layer_3 - start_layer_3))
        query_stats["time_layer3_total"] = end_layer_3 - start_layer_3

        return final_grouped_matches, query_stats

    def generateHTMLHeader(self):
        html = """
        <!DOCTYPE html>
        <HTML>
            <HEAD>
                <TITLE>Query""" + self.id + """</TITLE>
                <style>
                 #body {
                    margin: 0;
                    font-family: "Helvetica Neue";
                    font-size: 1em;
                    color: #222;
                    padding: 30px 60px;
                }

                #statsline {
                    font-size: 1.25em;
                }

                #logo {
                    width: 140px;
                    font-family: "Helvetica Neue";
                    font-weight: 250;
                    font-size: 2em;
                    float: left;
                }

                .score {
                    color: #999;
                }

                #queryblock {
                }

                #header {
                    background-color: #eee;
                    padding: 12px 30px;
                    overflow: auto;
                    border-bottom: 1px solid #ccc;
                }

                .searchbutton {
                    padding: 5px;
                    background-color: #efe;
                    border: 2px solid black;
                    height: 35px;
                    width: 60px;
                    font-size: 12pt;
                }

                .math_query img {
                    border: 2px solid black;
                    margin: 5px;
                    width: auto;
                    max-width: 500px;
                }

                .query_match_container img {
                    width: auto;
                    max-width: 500px;                    
                }

                .candidate_match_container img {
                    width: auto;
                    max-width: 500px;
                }

                .result_row_cell {
                    border-bottom: 2px solid #222222;
                    padding: 10px;
                }

                .result_row_cell b {
                    color: #00486A;
                }

                .query_match_container {
                    border-right: 2px dashed #444444;
                    padding-right: 10px;
                }

                .candidate_match_container {
                    margin-left: 10px;
                    padding-left: 10px;
                }

                .query_stats {
                    background-color: #eee;
                    padding: 12px 30px;
                    overflow: auto;
                    border-top: 1px solid #ccc;
                }

                .query_stats .layer_title {
                    color: #117711;
                    font-weight: bold;
                    text-align: center;
                }

                .query_stats .time {
                    color: #2277AA;
                    font-weight: bold;
                }

                .query_stats .field_name {
                    color: #052244;
                    text-align: right;
                    font-weight: bold;
                }

                .query_stats .layer_cell {
                    padding-left: 50px;
                    padding-right: 50px;
                    vertical-align: top;
                }

                </style>
                <meta charset=\"UTF-8\">
                <script type=\"text/javascript\" >
                    function hide_class(hide, class_name, second_class){
                        var nodes = document.getElementsByClassName(class_name);
                        var i;
                        for (i = 0; i < nodes.length; i++){
                            if (hide){
                                nodes[i].style.display = "none";
                            } else {
                                nodes[i].style.display = "";
                            }
                        }

                        //show the second
                        var nodes = document.getElementsByClassName(second_class);
                        for (i = 0; i < nodes.length; i++){
                            if (hide){
                                nodes[i].style.display = "";
                            } else {
                                nodes[i].style.display = "none";
                            }
                        }

                        if (hide){
                            document.getElementById('id_' + class_name + '_show').style.display = '';
                            document.getElementById('id_' + class_name + '_hide').style.display = 'none';
                        } else {
                            document.getElementById('id_' + class_name + '_show').style.display = 'none';
                            document.getElementById('id_' + class_name + '_hide').style.display = '';
                        }
                    }
                </script>
            </HEAD>
            <BODY>
                <font face=helvetica>
        """
        return html

    def get_html_queryblock(self, img_filenames):

        query_img = ""
        for img_filename in img_filenames:
            query_img += "<IMG src=\"{0:s}\" alt=\"{1:s}\">".format(img_filename, self.id)

        html_queryblock = """
        <!-- Query -->
        <div id="queryblock" align="left">
            <h2 align="left" >Search Results for: {0:s} ({1:s})</h2>
            <table>
                <tr><td>
                    <div class=\"math_query\">{2:s}</div>
                </td></tr>
            </table>
        </div>
        """.format(self.id, self.query_set.id, query_img)

        return html_queryblock

    def get_html_logo(self, img_filenames):
        prev_id = self.query_set.prev_query(self.id).id
        next_id = self.query_set.next_query(self.id).id

        result = """
        <div id="header">
            <table width="100%"><tr>
                <td width="180px">
                    <!-- Logo and buttons -->
                    <div>
                        <div id="logo">Tangent-V<br>
                            <table align="left" ><tr>
                                <td>
                                    <input type="button" id="id_match_debug_show" value="Show Match" onclick="hide_class(false, 'match_debug', 'match_normal');">
                                    <input type="button" id="id_match_debug_hide" value="Hide Match" style="background:yellow; display: none;"
                                        onclick="hide_class(true, 'match_debug', 'match_normal');">
                                </td>
                            </tr><tr>
                                <td>                                
                                    <table><tr>
                                        <td>
                                            <form action="../{0:s}/{0:s}.html">
                                                <button class="searchbutton" type="submit">Prev</button>    
                                            </form>
                                        </td><td>
                                            <form action="../{1:s}/{1:s}.html">
                                                <button class="searchbutton" type="submit">Next</button>
                                            </form>
                                        </td>
                                    </tr></table>
                                </td>
                            </tr></table>
                        </div>
                    </div>
                <td>{2:s}</td>
                <td align="right">
                    <!-- Search Button -->
                    <!-- <button id="searchbutton" type="button">Search</button> -->                    
                </td>
            </tr></table>
        </div>
        """.format(prev_id, next_id, self.get_html_queryblock(img_filenames))

        return result

    def generateHTMLFooter(self):
        html = """
                </font>
            </BODY>
        </HTML>
        """

        return html

    def generateHTMLResults(self, results_info):
        html_header = "<TABLE class=\"results_section\" align=\"center\" >"
        html_footer = "</TABLE>"

        html_rows = []
        for idx, match_info in enumerate(results_info):

            locations_title_str = "Locations: "
            locations_str = ", ".join([str(location) for location in match_info["locations"]])
            match_link = match_info["match_link"]

            if isinstance(match_info["score"], list):
                top_score = match_info["score"][0]
                middle_row = """
                        <tr>
                            <TD class=\"query_match_container\"><b>Precision: </b>{0:.4f}%</td>
                            <TD class=\"candidate_match_container\"><br /></td>
                        </tr>""".format(match_info["score"][1] * 100.0)
            else:
                top_score = match_info["score"]
                middle_row = ""

            row_html = """
            <TR class="result_row" >
                <TD class="result_row_cell">
                    <table>
                        <tr>
                           <TD class=\"query_match_container\"><b>Query Match #{9:d}</b></td>
                           <TD class=\"candidate_match_container\"><b>Match on: </b>{0:s}</td>
                        </tr>
                        <tr>
                            <TD class=\"query_match_container\"><b>Score: </b>{1:.4f}%</td>
                            <TD class=\"candidate_match_container\"><b>{2:s}: </b>{3:s}</td>
                        </tr>
                        {10:s}
                        <tr>
                            <td class=\"query_match_container\">
                                <img src="{4:s}" class=\"match_normal\" >
                                <img src="{5:s}" class=\"match_debug\" style=\"display: None;\" >
                            </td>
                            <td class=\"candidate_match_container\">
                                <a href="{6:s}" target="_blank">
                                    <img src="{7:s}" class=\"match_normal\" >
                                    <img src="{8:s}" class=\"match_debug\" style=\"display: None;\" >
                                </a>
                            </td>
                        </tr>
                    </table>
                </TD>
            </TR>
            """.format(match_info["source"], top_score * 100.0, locations_title_str, locations_str,
                       match_info["normal_query_image"], match_info["debug_query_image"], match_link,
                       match_info["normal_candidate_image"], match_info["debug_candidate_image"], idx + 1, middle_row)

            html_rows.append(row_html)

        html_content = "".join(html_rows)

        return html_header + html_content + html_footer

    def valOrDef(self, dictionary, value_name, default):
        if value_name in dictionary:
            return dictionary[value_name]
        else:
            return default

    def generateHTMLStats(self, query_stats):

        total_time = query_stats['time_layer1_total'] + query_stats['time_layer2_total'] + query_stats['time_layer3_total']
        html_global = """
        <p>
		    <b class="field_name">Total Query Execution Time: </b><b class="time">{0:s}</b><br />
		</p>""".format(TimeHelper.secondsToStr(total_time))

        html_layer_1 = """
        <table>
            <tr>
                <td colspan="3" class="layer_title">Layer 1 (Index Search) Stats</td>
            </tr>
            <tr>
                <td class="field_name">Total Time: </td>
                <td class="time">{0:s}</td>
                <td><br /></td>
            </tr>
            <tr>
                <td class="field_name">Query Pairs: </td>
                <td>{1:d}</td>
                <td><br /></td>
            </tr>
            <tr>
                <td class="field_name">Matching Raw Pairs: </td>
                <td>{2:d}</td>
                <td class="time">{3:s}</td>
            </tr>
            <tr>
                <td class="field_name">Image Pairs: </td>
                <td>{4:d}</td>
                <td class="time">{5:s}</td>
            </tr>
            <tr>
                <td class="field_name">Initial Unique Images: </td>
                <td>{6:d}</td>
                <td class="time">{7:s}</td>
            </tr>
            <tr>
                <td class="field_name">Final Unique Images : </td>
                <td>{8:d}</td>
                <td class="time">{9:s}</td>
            </tr>
        </table>""".format(TimeHelper.secondsToStr(query_stats['time_layer1_total']),
                           query_stats['count_layer1_query_pairs'], query_stats['count_layer1_raw_matches'],
                           TimeHelper.secondsToStr(query_stats['time_layer1_list_raw']),
                           query_stats['count_layer1_aggregated_matches'],
                           TimeHelper.secondsToStr(query_stats['time_layer1_list_aggregate']),
                           query_stats['count_layer1_initial_images'],
                           TimeHelper.secondsToStr(query_stats['time_layer1_pre_ranking']),
                           query_stats['count_layer1_images_matched'],
                           TimeHelper.secondsToStr(query_stats['time_layer1_list_merge']))

        html_layer_2 = """
        <table>
            <tr>
                <td colspan="3" class="layer_title">Layer 2 (Match Growing) Stats</td>
            </tr>
            <tr>
                <td class="field_name">Total Time: </td>
                <td class="time">{0:s}</td>
                <td><br /></td>
            </tr>
            <tr>
                <td class="field_name">Initial Matches: </td>
                <td>{1:d}</td>
                <td><br /></td>
            </tr>
            <tr>
                <td class="field_name">Grown Matches: </td>
                <td>{2:d}</td>
                <td class="time">{3:s}</td>
            </tr>
            <tr>
                <td class="field_name">Connected Matches: </td>
                <td>{4:d}</td>
                <td class="time">{5:s}</td>
            </tr>
            <tr>
                <td class="field_name">Filtered Matches: </td>
                <td>{6:d}</td>
                <td class="time">{7:s}</td>
            </tr>
            <tr>
                <td class="field_name">Joined Matches: </td>
                <td>{8:d}</td>
                <td class="time">{9:s}</td>
            </tr>
        </table>
        """.format(TimeHelper.secondsToStr(query_stats['time_layer2_total']),
                   query_stats['count_layer2_initial_matches'], query_stats['count_layer2_grown_matches'],
                   TimeHelper.secondsToStr(query_stats['time_layer2_grow_merge']),
                   query_stats['count_layer2_connected_matches'],
                   TimeHelper.secondsToStr(query_stats['time_layer2_grow_connect']),
                   query_stats['count_layer2_final_matches'],
                   TimeHelper.secondsToStr(query_stats['time_layer2_match_filter']),
                   query_stats['count_layer2_join_disconnected'],
                   TimeHelper.secondsToStr(query_stats['time_layer2_join_disconnected']))

        html_layer_3 = """
        <table>
            <tr>
                <td colspan="3" class="layer_title">Layer 3 (Match Grouping) Stats</td>
            </tr>
            <tr>
                <td class="field_name">Total Time: </td>
                <td class="time">{0:s}</td>
                <td><br /></td>
            </tr>
            <tr>
                <td class="field_name">Initial Groups: </td>
                <td>{1:d}</td>
                <td class="time"><br /></td>
            </tr>
            <tr>
                <td class="field_name">Final Groups: </td>
                <td>{2:d}</td>
                <td class="time"><br /></td>
            </tr>
        </table>
        """.format(TimeHelper.secondsToStr(query_stats['time_layer3_total']),
                   query_stats['count_layer3_initial_matches'], query_stats['count_layer3_final_matches'])

        html_content = """
        <div class="query_stats">
			<table align="center">
			    <tr>
			        <td class="layer_cell">{0:s}</td>
			        <td class="layer_cell">{1:s}</td>
			        <td class="layer_cell">{2:s}</td>
			    </tr>
			</table>
		</div>""".format(html_layer_1, html_layer_2, html_layer_3)

        html_stats = html_global + html_content
        return "<div class=\"query_stats\">" + html_stats + "</div>"

    def SaveResultHTML(self, output_dir, image_dir, image_index, scored_groups, query_stats, top_K, link_generator,
                       match_padding=None):
        # assume input group is a list of tuple in the format
        # [ ...., (best_score, best_score_kf, match_group), ... ]

        # first, create the output directory for images ...
        query_prefix = output_dir + "/" + self.id

        if not os.path.exists(query_prefix):
            os.mkdir(query_prefix)

        if not os.path.exists(query_prefix + "/images"):
            os.mkdir(query_prefix + "/images")

        # save first the query image ...
        query_image_filenames = []
        for idx, query_image in enumerate(self.query_images):
            image_filename = "images/query_" + str(idx) + ".png"
            query_image_filenames.append(image_filename)

            # save image to file ...
            base, ext = os.path.splitext(query_image.image_url)
            if ext.lower() == ".pdf":
                # input is PDF ... convert to PNG ...
                image = QueryInfo.pdf2png(query_image.image_url, query_prefix + "/images/query_" + str(idx))

                # resize to graph size ...
                new_size = query_image.graph.width, query_image.graph.height
                image = cv2.resize(image, new_size)

                # assign rendered image back to the graph ..
                # invert y-axis and
                # use green channel which should be dark on wildcards
                binary_query_img = image[::-1, :, 1]
                query_ccs = RecognitionHelper.CCsFromBinary(255 - binary_query_img)
                query_image.graph.approximate_recover_for_rendering(query_ccs)
            else:
                # treat as image by default ...
                cv2.imwrite(query_prefix + "/" + image_filename, query_image.raw_image)

        results_info = []
        for idx, (best_score, best_score_img, match_group) in enumerate(scored_groups[:top_K]):
            query_image_filename = "images/result_query_match_{0:s}_{1:d}.png"
            candidate_image_filename = "images/result_candidate_match_{0:s}_{1:d}.png"

            best_match = match_group.matches_per_image[best_score_img]
            best_region = match_group.regions_per_image[best_score_img]
            assert isinstance(best_match, GraphMatch)

            # print("\n")
            # print(best_match)
            # best_match.print_rec_confidences()

            # TODO: best_score should be a list in all cases, this would simplify code everywhere
            if isinstance(best_score, list):
                top_score = best_score[0]
            else:
                top_score = best_score

            if IndexParameters.SimilarityMetric == IndexParameters.Metric_ProductBased:
                normalized_best = top_score / best_match.query_graph.max_prod_pairs_weight
            elif IndexParameters.SimilarityMetric == IndexParameters.Metric_MinBased:
                normalized_best = top_score / best_match.query_graph.max_comb_pairs_weight
            else:
                normalized_best = top_score

            if isinstance(best_score, list):
                normalized_best = [normalized_best] + best_score[1:]

            img_id = best_score_img
            index_image = image_index.images[img_id]
            base, ext = os.path.splitext(index_image.file_path)
            if ext.lower() == ".pdf":
                # special case: pdf ... convert to PNG ..
                print(image_dir + "/" + index_image.file_path)
                image = QueryInfo.pdf2png(image_dir + "/" + index_image.file_path, query_prefix + "/images/temporal")

                # resize to graph size ...
                new_size = best_match.candidate_graph.width, best_match.candidate_graph.height
                image = cv2.resize(image, new_size)

                # invert y-axis ...
                image = image[::-1, :, :]
                use_approx_recovery = True
            else:
                # treat as image, simply read from file ...
                image = cv2.imread(image_dir + "/" + index_image.file_path)
                use_approx_recovery = False

            binary_image = image[:, :, 0]

            # recover missing data ...
            candidate_cc = RecognitionHelper.CCsFromBinary(255 - binary_image)
            if use_approx_recovery:
                # recover CCs based on approximate bounding box matching ...
                best_match.candidate_graph.approximate_recover_for_rendering(candidate_cc)
            else:
                # normal recovery based on exact bounding box matching
                best_match.candidate_graph.recover_for_rendering(candidate_cc)

            h, w, _ = image.shape
            min_x, max_x, min_y, max_y = best_region

            if match_padding is None:
                # use the entire image ...
                cut_min_x = 0
                cut_max_x = w
                cut_min_y = 0
                cut_max_y = h
            else:
                # use a cut version of the image ...
                padding_top, padding_right, padding_bottom, padding_left =  match_padding
                cut_min_x = max(0, min_x - padding_left)
                cut_max_x = min(w, max_x + padding_right)
                cut_min_y = max(0, min_y - padding_top)
                cut_max_y = min(h, max_y + padding_bottom)

            # generate normal view ....
            query_match, candidate_match = best_match.getImage(True, False, False)
            normal_image = binary_image[cut_min_y:cut_max_y, cut_min_x:cut_max_x]

            if ext.lower() == ".pdf":
                query_match = query_match[::-1, :, :]
                normal_image = normal_image[::-1, :]

            q_normal_img_file = query_image_filename.format("normal", idx)
            c_normal_img_file = candidate_image_filename.format("normal", idx)
            cv2.imwrite(query_prefix + "/" + q_normal_img_file, query_match)
            cv2.imwrite(query_prefix + "/" + c_normal_img_file, normal_image)

            # generate debug view ...
            query_match, candidate_match = best_match.getImage(True, True, True)
            debug_keyframe = candidate_match[cut_min_y:cut_max_y, cut_min_x:cut_max_x]

            if ext.lower() == ".pdf":
                query_match = query_match[::-1, :, :]
                debug_keyframe = debug_keyframe[::-1, :, :]

            q_debug_img_file = query_image_filename.format("debug", idx)
            c_debug_img_file = candidate_image_filename.format("debug", idx)
            cv2.imwrite(query_prefix + "/" + q_debug_img_file, query_match)
            cv2.imwrite(query_prefix + "/" + c_debug_img_file, debug_keyframe)

            if link_generator is None:
                match_link = ""
            else:
                match_link = link_generator.generate_match_link(image_index, match_group)

            match_info = {
                # "lecture": kf_index.lectures[match_group.lecture_id],
                # "keyframes": sorted(list(match_group.matches_per_keyframe.keys())),
                "normal_query_image": q_normal_img_file,
                "normal_candidate_image": c_normal_img_file,
                "debug_query_image": q_debug_img_file,
                "debug_candidate_image": c_debug_img_file,
                "score": normalized_best,
                "locations": index_image.source_ids,
                "source": index_image.file_path,
                "match_link": match_link,
            }

            results_info.append(match_info)


        # now, generate images and HTML ....
        header = self.generateHTMLHeader()
        header += self.get_html_logo(query_image_filenames)
        content = self.generateHTMLResults(results_info) + self.generateHTMLStats(query_stats)
        footer = self.generateHTMLFooter()

        # save HTML ...
        HTML = header + content + footer
        with open(query_prefix + "/" + self.id + ".html", "w", encoding="utf") as out_file:
            out_file.write(HTML)

    def getResultsTREC(self, image_index, scored_groups, top_K):
        TREC_lines = []
        # Query id, 1, location id, rank, score, set id
        line_format = self.id + "\t1\t{0:s}\t{1:d}\t{2:.8f}\t" + self.query_set.id + "\n"
        rank = 1
        # TREC format does not allow duplicated ids, only keep the best match from each location ...
        included_ids = {}
        for idx, (best_score, best_score_img, match_group) in enumerate(scored_groups[:top_K]):
            best_match = match_group.matches_per_image[best_score_img]
            assert isinstance(best_match, GraphMatch)

            if isinstance(best_score, list):
                if len(best_score) > 1:
                    # use reciprocal rank to keep tie breakers ...
                    normalized_best = 1 / (idx + 1)
                else:
                    normalized_best = best_score[0]
            else:
                normalized_best = best_score

            img_id = best_score_img
            index_image = image_index.images[img_id]

            # each source is a matched document ...
            for source_id in index_image.source_ids:
                if not source_id in included_ids:
                    included_ids[source_id] = True
                    source_line = line_format.format(source_id, rank, normalized_best)
                    TREC_lines.append(source_line)
                    rank += 1

        return TREC_lines

    def getResultsBBox(self, image_index, scored_groups, top_K):
        BBox_lines = []
        # Query id, location id, x, y, w, h, rank, score(s)
        line_format = self.id + ",{0:s},{1:d},{2:d},{3:d},{4:d},{5:d}"
        rank = 1

        for idx, (best_score, best_score_img, match_group) in enumerate(scored_groups[:top_K]):
            best_match = match_group.matches_per_image[best_score_img]
            assert isinstance(best_match, GraphMatch)

            # get match box ...
            best_region = match_group.regions_per_image[best_score_img]
            min_x, max_x, min_y, max_y = best_region
            w = max_x - min_x + 1
            h = max_y - min_y + 1

            img_id = best_score_img
            index_image = image_index.images[img_id]

            # each source is a matched document ...
            for source_id in index_image.source_ids:
                source_line = line_format.format(source_id, min_x, min_y, w, h, rank)

                if isinstance(best_score, list):
                    source_line += "".join([",{0:.4f}".format(score) for score in best_score]) + "\n"
                else:
                    source_line += ",{0:.4f}\n".format(best_score)

                BBox_lines.append(source_line)
                rank += 1

        return BBox_lines

    @staticmethod
    def pdf2png(input_pdf_file, output_prefix, dpi=600, crop_content=True, crop_rewrite=True):
        # This function assumes that the pdftoppm tool is available
        # it should be on linux distributions. However, windows will need to install poppler
        # NOTE: it seems that some versions of poppler do not support the png flag directly.
        #       for increased compatibility, JPEG is used first and then converted to PNG
        convert_command = "pdftoppm -singlefile -r {0:d} -jpeg {1:s} {2:s}"
        os.system(convert_command.format(dpi, input_pdf_file, output_prefix))

        # convert ...
        # ...read ...
        tempo_img = cv2.imread(output_prefix + ".jpg")
        # ... binarize (remove compression artifacts) ...
        tempo_img = (tempo_img > 128).astype(np.uint8) * 255
        # ... save as png ..
        cv2.imwrite(output_prefix + ".png", tempo_img)
        # ... remove temporal jpeg ...
        os.remove(output_prefix + ".jpg")


        if crop_content:
            # read the output file ...
            input_image = cv2.imread(output_prefix + ".png")

            # get profile-based boundaries ...
            comb_img = 255 - input_image
            # cv2.imshow("tempo", comb_img)
            # cv2.waitKey()

            h_prof = comb_img.sum(axis=0).nonzero()
            x_min = h_prof[0].min()
            x_max = h_prof[0].max()

            v_prof = comb_img.sum(axis=1).nonzero()
            y_min = v_prof[0].min()
            y_max = v_prof[0].max()

            # print(input_image.shape)
            # print((x_min, x_max, y_min, y_max))

            # crop ...
            cropped_img = input_image[y_min:y_max + 1, x_min:x_max + 1, :]

            # save
            if crop_rewrite:
                cv2.imwrite(output_prefix + ".png", cropped_img)

            return cropped_img
        else:
            return None

    @staticmethod
    def FromXMLNode(query_set, xml_path, query_xml):
        namespace = QueryInfo.XMLNamespace

        # basic query info
        query_id = query_xml.find(namespace + 'Id').text

        query_type_str = query_xml.find(namespace + 'Type').text.lower()
        if query_type_str == "single-image":
            query_type = QueryInfo.TypeSingleImage
        else:
            raise Exception("Unknown query type: " + query_type_str)

        group_xml = query_xml.find(namespace + 'Groups')
        if group_xml is not None:
            group_mode_str = group_xml.find(namespace + 'Mode').text.lower()
        else:
            group_mode_str = ""

        group_options = {}
        if group_mode_str == "image":
            group_mode = QueryInfo.GroupModeImage
        elif group_mode_str == "sequence":
            group_mode = QueryInfo.GroupModeSequence
            group_options["group_by"] = group_xml.find(namespace + 'GroupBy').text
            group_options["sequence_by"] = group_xml.find(namespace + 'SequenceBy').text
        elif group_mode_str == "attribute":
            group_mode = QueryInfo.GroupModeAttribute
            # check for group-by attributes (there can be more than one!)
            group_options["group_by"] = []
            for group_by_xml in group_xml.findall(namespace + 'GroupBy'):
                group_options["group_by"].append(group_by_xml.text)
        else:
            group_mode = QueryInfo.GroupModeNone

        # create and add query
        query = QueryInfo(query_id, query_set, query_type, group_mode, group_options)

        # Load query image(s)
        images_root = query_xml.find(namespace + 'Images')
        all_xml_images = images_root.findall(namespace + 'Image')
        for xml_image in all_xml_images:
            # load from XML ...
            query_image = QueryImage.FromXMLNode(xml_image, query_id, namespace)
            # correct image path to current path ...
            query_image.image_url = xml_path + "/" + query_image.image_url
            # add to query images ...
            query.query_images.append(query_image)

        return query
