
import os
import sys
import codecs
import shutil

from bs4 import BeautifulSoup

from concurrent.futures import ProcessPoolExecutor

# if sys.stdout.encoding != 'utf8':
#    sys.stdout = codecs.getwriter('utf8')(sys.stdout.buffer, 'strict')
# if sys.stderr.encoding != 'utf8':
#     sys.stderr = codecs.getwriter('utf8')(sys.stderr.buffer, 'strict')


class MathScraperHelper:
    def __init__(self, mathscraper_path):
        self.math_scraper_path = mathscraper_path
        self.batch_command = "java -jar {0:s} -batch -p {1:s} {1:s}"
        self.file_command = "java -jar {0:} -p {1:s}"

    def extract_compound_labels(self, compound_char):
        return [char.text for char in compound_char.find_all("char")]

    def extract_compound_bboxes(self, compound_char):
        all_x, all_y, all_max_x, all_max_y = [], [], [], []
        for inner_char in compound_char.find_all("char"):
            x, y, w, h = [float(val) for val in inner_char["bbox"].split(" ")]

            all_x.append(x)
            all_y.append(y)
            all_max_x.append(x + w)
            all_max_y.append(y + h)

        x = min(all_x)
        y = min(all_y)
        w = max(all_max_x) - x
        h = max(all_max_y) - y

        return x, y, w, h

    def extract_compound_data(self, compound_char, recompute_bbox):
        if not recompute_bbox:
            # use provided bbox data
            x, y, w, h = [float(val) for val in compound_char["bbox"].split(" ")]
        else:
            # re-compute bounding box from inner elements
            x, y, w, h = self.extract_compound_bboxes(compound_char)

        # ... here label is included as an attribute ...
        label = compound_char["label"]

        if label == "unknown" or label == "":
            # no label is provided, compute a set of labels from inner chars
            inner_labels = self.extract_compound_labels(compound_char)

            label_conf = {}
            for label in inner_labels:
                if label in label_conf:
                    label_conf[label] += 1
                else:
                    label_conf[label] = 1

            # multi-labels, each with its own confidence
            for label in label_conf:
                label_conf[label] /= len(inner_labels)
        else:
            # use single label provided
            label_conf = {label: 1.0}

        return x, y, w, h, label_conf

    def symbols_from_xml_file(self, input_filename):
        # read input file ...
        with open(input_filename, "r", encoding="utf8") as input_file:
            soup = BeautifulSoup(input_file, "lxml")

        file_data = []

        # start with compound characters ...
        all_compound_chars = soup.find_all("compoundchar")

        unique_compound_groups = {}
        for compound_char in all_compound_chars:
            # remove from parent ...
            compound_char.extract()

            compound_id = compound_char["id"]
            if not compound_id in unique_compound_groups:
                unique_compound_groups[compound_id] = [compound_char]
            else:
                unique_compound_groups[compound_id].append(compound_char)

        for compound_id in unique_compound_groups:
            if len(unique_compound_groups[compound_id]) == 1:
                # no split compound character ...
                compound_char = unique_compound_groups[compound_id][0]

                x, y, w, h, label_conf = self.extract_compound_data(compound_char, False)
                file_data.append((x, y, w, h, label_conf))
            else:
                # should just split each unit as a separate compound char? (simpler?)
                # -> Each separated compound needs to have its BBox recomputed!
                for compound_char in unique_compound_groups[compound_id]:
                    # force bbox re-computation for each piece of compound char
                    x, y, w, h, label_conf = self.extract_compound_data(compound_char, True)
                    file_data.append((x, y, w, h, label_conf))

        all_chars = soup.find_all("char")

        # ... now the regular characters ...
        for ext_char in all_chars:
            # data ...
            x, y, w, h = [float(val) for val in ext_char["bbox"].split(" ")]
            # ... label is within inner text ...
            label = ext_char.text
            file_data.append((x, y, w, h, {label: 1.0}))

        return file_data

    def symbols_from_pdf(self, input_filename):
        # Java won't take paths with mixed "\" and "/" on windows
        input_filename = input_filename.replace("/", os.path.sep)

        file_dir, suffix_name = os.path.split(input_filename)
        basename, ext = os.path.splitext(suffix_name)

        temporal_dir = file_dir + "/Output"
        delete_temporal = not os.path.isdir(temporal_dir)
        temporal_file = temporal_dir + "/" + basename + ".xml"

        # self.file_command = "java -jar {0:} -p {1:s}"
        command = self.file_command.format(self.math_scraper_path, input_filename)
        # execute MathScrapper
        os.system(command)

        # print((input_filename, file_dir, suffix_name, basename, ext, temporal_dir, delete_temporal, temporal_file))
        # print(command)

        file_data = self.symbols_from_xml_file(temporal_file)

        if delete_temporal:
            # delete complete temporary output
            shutil.rmtree(temporal_dir)
        else:
            # only delete temporary xml file ...
            os.remove(temporal_file)

        return file_data

    def symbols_from_directory(self, input_directory, verbose=True):
        if verbose:
            print("Processing: " + input_directory, flush=True)

        elements = os.listdir(input_directory)

        # execute the MathScrapper tool ...
        os.system(self.batch_command.format(self.math_scraper_path, input_directory))

        all_dir_data = []
        for element in elements:
            base, ext = os.path.splitext(element)

            if ext.lower() == ".pdf":
                # expecting an XML to be present for this pdf ...
                pdf_filename = input_directory + "/" + element
                xml_filename = input_directory + "/" + base + ".xml"

                # read XML ...
                print("Try to load: " + xml_filename)
                if os.path.exists(xml_filename):
                    # load it ...
                    file_symbols = self.symbols_from_xml_file(xml_filename)
                    # add to dir data
                    all_dir_data.append((pdf_filename, file_symbols))

                    # delete it ...
                    os.remove(xml_filename)
                else:
                    print("-> WARNING: " + xml_filename + " not found!!!", flush=True)

        return input_directory, all_dir_data

    def __rec_extract_pdf_dirs(self, input_directory):
        elements = os.listdir(input_directory)

        collected_files = {}
        for element in elements:
            element_path = input_directory + "/" + element

            if os.path.isdir(element_path):
                print("...Scanning: " + element_path, flush=True, end="\r")
                collected_files.update(self.__rec_extract_pdf_dirs(element_path))
            else:
                base, ext = os.path.splitext(element)

                if ext == ".pdf":
                    if not input_directory in collected_files:
                        collected_files[input_directory] = []

                    collected_files[input_directory].append(element_path)

        return collected_files

    def scan_pdf_dirs(self, input_directory):
        return self.__rec_extract_pdf_dirs(input_directory)

    def extract_data(self, input_directory, batch_workers, verbose=True):
        if verbose:
            print("... scanning input files ... ", flush=True)
        collected_files = self.scan_pdf_dirs(input_directory)
        total_files = sum([len(collected_files[dir_path]) for dir_path in collected_files])

        if verbose:
            message = "\n\nA total of {0:d} files from {1:d} directories will be processed"
            print(message.format(total_files, len(collected_files)))

        all_data = []
        total_symbols = 0
        total_files_proc = 0
        with ProcessPoolExecutor(max_workers=batch_workers) as executor:
            pdf_directories = sorted(list(collected_files.keys()))
            # for idx, (directory, all_files_data) in enumerate(map(symbols_from_directory, pdf_directories)):
            for idx, (directory, all_files_data) in enumerate(executor.map(self.symbols_from_directory,
                                                                           pdf_directories)):
                for filename, file_data in all_files_data:
                    total_symbols += len(file_data)

                    # remove common input image dir
                    filename = filename[len(input_directory) + 1:]
                    all_data.append((filename, file_data))

                total_files_proc += len(all_files_data)

                if verbose:
                    print("A total of {0:d} files processed so far".format(total_files_proc), flush=True)

        if verbose:
            print("A total of {0:d} symbols from {1:d} files were extracted".format(total_symbols, len(all_data)))

        return all_data, (total_symbols, total_files_proc)

