
import pickle
import numpy as np
from sklearn.preprocessing import StandardScaler

class FeatureDataset:
    def __init__(self, data, labels, label_mapping, class_mapping, sources=None, config=None):
        # main data
        self.data = data
        self.labels = labels
        self.label_mapping = label_mapping
        self.class_mapping = class_mapping

        # optional ...
        self.sources = sources
        self.config = config
        self.normalizer = None

    def normalize_data(self, normalizer=None):
        if normalizer is None:
            # train new normalizer
            self.normalizer = StandardScaler()
            self.data = self.normalizer.fit_transform(self.data)
        else:
            # use the provided normalizer
            self.normalizer = normalizer
            self.data = self.normalizer.transform(self.data)


    def save_to_file(self, filename):
        out_file = open(filename, 'wb')
        pickle.dump(self, out_file, pickle.HIGHEST_PROTOCOL)
        out_file.close()

    def num_samples(self):
        return self.data.shape[0]

    def num_attributes(self):
        return self.data.shape[1]

    def num_classes(self):
        return len(self.label_mapping)

    def split_data(self, n_partitions):
        by_class = np.argsort(self.labels, axis=0)

        partitions_data, partitions_labels, partitions_indices = [], [], []

        for i in range(n_partitions):
            partition_indices = np.squeeze(by_class[i::n_partitions])

            partitions_data.append(self.data[partition_indices])
            partitions_labels.append(self.labels[partition_indices])
            partitions_indices.append(partition_indices)

        return partitions_data, partitions_labels, partitions_indices

    @staticmethod
    def from_INKML_data(raw_features, raw_labels, label_mapping=None, class_mapping=None, sources=None, config=None):
        # create data
        data = np.array(raw_features)

        # create label mapping, and map raw labels
        mapped_labels, label_mapping, class_mapping = FeatureDataset.create_update_label_mapping(raw_labels, label_mapping, class_mapping)

        return FeatureDataset(data, mapped_labels, label_mapping, class_mapping, sources, config)

    @staticmethod
    def create_update_label_mapping(raw_labels, label_mapping=None, class_mapping=None):
        if label_mapping is None:
            modify_mapping = True
            label_mapping = []
            class_mapping = {}
        else:
            modify_mapping = False

        mapped_labels = np.zeros((len(raw_labels), 1), dtype = np.int32)
        for idx, class_label in enumerate(raw_labels):
            if not class_label in class_mapping:
                class_mapping[class_label] = len(label_mapping)
                label_mapping.append(class_label)
                if not modify_mapping:
                    print("Warning: new class <" + class_label + "> added to existing mapping")

            mapped_labels[idx, 0] = class_mapping[class_label]

        return mapped_labels, label_mapping, class_mapping

    @staticmethod
    def load_from_file(filename):
        in_file = open(filename, 'rb')
        dataset = pickle.load(in_file)
        in_file.close()

        return dataset